''' Load Fritzing part files as Schemdraw elements

    See https://github.com/fritzing/fritzing-app/wiki/2.1-Part-file-format

    Only extracts the breadboard view (schematic, PCB, and icon are not used)
'''
from __future__ import annotations
from typing import Optional
import re
import warnings
import zipfile
from io import BytesIO
from collections import namedtuple
from xml.etree import ElementTree as ET

from ..util import Point
from ..backends import matrix
from ..elements import ElementImage
from .pictorial import parse_size_to_units


FritzingInfo = namedtuple('FritzingInfo', 'author version title url label date description')


def find_transforms(element, tree):
    '''' Find all 'transform' tags in tree leading up to element '''
    parents = {c: p for p in tree.iter() for c in p}
    xforms = []
    def _find(elm):
        if (xf := elm.get('transform')):
            xforms.append(xf)
        p = parents.get(elm)
        if p is not None:
            _find(p)
    _find(element)
    return xforms


def extract(elm: ET.Element, tag: str) -> Optional[str]:
    ''' Extract the tag's text from the element, returning None if tag 
        doesn't exist
    '''
    item = elm.find(tag)
    return item.text if item is not None else None


def fritz_parts(fname: str) -> list[str | None]:
    ''' List titles of all Fritzing parts in the file '''
    zip = zipfile.ZipFile(fname)
    parts = [f.filename for f in zip.infolist() if f.filename.endswith('.fzp')]
    names = []
    for part in parts:    
        module = ET.fromstring(zip.read(part))
        title = module.find('title')
        if title is not None:
            names.append(title.text)
    return names


class FritzingPart(ElementImage):
    ''' Load a Fritzing Part File as a Schemdraw Element

        Anchors will be extracted from the Part definition file.
        Note some anchors are not valid Python identifiers and
        therefore must be accessed through the Element.absanchors
        dictionary rather than an attribute of the element instance.

        Args:
            fname: Filename of fritzing .fzpz archive
            partname: Name of part within the file. Use `listparts` to
                show all names. If not provided, first part is drawn.
            partidx: Index of part within the file. If not provided,
                first part is drawn. Overrides `partname`.
            scale: Scale factor
    '''
    def __init__(self, fname: str,
                 partname: Optional[str] = None,
                 partidx: Optional[int] = None,
                 scale: float = 1.0):
        self.fname = fname
        self.zip = zipfile.ZipFile(self.fname)
        parts = [f.filename for f in self.zip.infolist() if f.filename.endswith('.fzp')]

        if partidx is not None:
            part = parts[partidx]
        elif partname is not None:
            part = parts[fritz_parts(fname).index(partname)]
        else:
            part = parts[0]

        self.module = ET.fromstring(self.zip.read(part))
        self.info = FritzingInfo(
            author=extract(self.module, 'author'),
            version=extract(self.module, 'version'),
            title=extract(self.module, 'title'),
            url=extract(self.module, 'url'),
            label=extract(self.module, 'label'),
            date=extract(self.module, 'date'),
            description=extract(self.module, 'description'),
        )

        views = self.module.find('views')
        if views is None:
            raise ValueError('Part has no views')

        breadboard = views.find('breadboardView')
        if breadboard is None:
            raise ValueError('Part has no breadboardView')

        layers = breadboard.find('layers')
        if layers is None:
            raise ValueError('Part breadboardView has no layers')

        svgfile = layers.get('image', '')
        image = self.zip.read('svg.' + svgfile.replace('/', '.'))
        imagebuf = BytesIO(image)
        self.imagexml = ET.fromstring(image)

        width = self.imagexml.get('width', '0')
        height = self.imagexml.get('height', '0')
        self.width_units = parse_size_to_units(width) * scale
        self.height_units = parse_size_to_units(height) * scale

        if self.width_units == 0 or self.height_units == 0:
            warnings.warn('Part has 0 width or height')

        viewbox = self.imagexml.get('viewBox')
        if viewbox:
            _, _, vieww, viewh = viewbox.split()
            self.vieww = float(vieww)
            self.viewh = float(viewh)
            self._scale = self.width_units / self.vieww
        else:
            # Some examples don't have a viewBox, assume same as width/height
            try:
                width_f = float(width)
            except ValueError:
                width_f = float(width[:-2])
            self._scale = self.width_units / width_f

        super().__init__(image=imagebuf, imgfmt='svg', width=self.width_units, height=self.height_units)
        self._findanchors()

    def _anchor_position(self, name: str, anchorelm: ET.Element) -> tuple[Optional[float], Optional[float]]:
        ''' Extract position of the anchor element from its SVG tag '''
        matrices: list[matrix.Matrix3x3] = []
        anchorx: Optional[float] = None
        anchory: Optional[float] = None

        # Anchor is SVG tag, pick out the exact point
        if anchorelm.tag.endswith('circle') or anchorelm.tag.endswith('ellipse'):
            anchorx = float(anchorelm.get('cx', '0'))
            anchory = float(anchorelm.get('cy', '0'))
        elif anchorelm.tag.endswith('rect'):
            rectwidth = float(anchorelm.get('width', '0'))
            rectheight = float(anchorelm.get('height', '0'))
            anchorx = float(anchorelm.get('x', '0')) + rectwidth/2
            anchory = float(anchorelm.get('y', '0')) + rectheight/2
        elif anchorelm.tag.endswith('polygon'):
            pointlist = anchorelm.get('points', '')
            pointstrs = re.split(',| ', pointlist)
            points = [float(p) for p in pointstrs if p]
            xpoints = points[::2]
            ypoints = points[1::2]
            anchorx = (max(xpoints) + min(xpoints)) / 2
            anchory = (max(ypoints) + min(ypoints)) / 2
        elif anchorelm.tag.endswith('path'):
            path = anchorelm.get('d', '')
            path = path.replace('-', ' -')  # Some negative numbers don't have space before
            pointstrs = re.split(',| |M|L|H|V|Q|C|A|Z|m|l|h|v|q|c|a|z', path)
            # First item in list is 'M' directive, or similar
            # Let anchor just be the first point in the path.
            # Could be that center of path was intended, but that will get weird
            # with Bezier control points, etc.
            # NOTE: relative (lowercase m) moves won't work correctly
            pointstrs = [p for p in pointstrs if p]
            anchorx = float(pointstrs[0])
            anchory = float(pointstrs[1])
        elif anchorelm.tag.endswith('g'):
            # Set anchor based on first element within the group
            if len(anchorelm) > 0:
                return self._anchor_position(name, anchorelm[0])
            else:
                warnings.warn(f'Connector {name} is empty')
                return None, None
        else:
            warnings.warn(f'Connector {name} unimplemented connector tag: {anchorelm.tag}')
            return None, None

        # Pick out SVG transforms leading up to the anchor element
        xforms = find_transforms(anchorelm, self.imagexml)
        for xf in xforms:
            for xfmode, value in re.findall(r'(.*?)\((.*?)\)', xf):
                xfmode = xfmode.strip()
                valuestrs = re.split(',| ', value)
                values = [float(v) for v in valuestrs if v]

                if xfmode == 'translate':
                    m = matrix.matrix_translate(*values)
                elif xfmode == 'rotate':
                    m = matrix.matrix_rotate(*values)
                elif xfmode == 'scale':
                    m = matrix.matrix_scale(*values)
                elif xfmode == 'skewX':
                    m = matrix.matrix_skewx(values[0])
                elif xfmode == 'skewY':
                    m = matrix.matrix_skewy(values[0])
                elif xfmode == 'matrix':
                    m = matrix.matrix(*values)
                else:
                    raise NotImplementedError(f'Transform {xfmode} not implemented')
                matrices.append(m)

        # Apply the transformation matrices and scale to drawing units
        anchorx, anchory = matrix.transform_all(Point((anchorx, anchory)), matrices)
        anchorx = anchorx * self._scale
        anchory = self.height_units - anchory * self._scale
        return anchorx, anchory

    def _findanchors(self):
        ''' Find anchor positions '''
        for connector in self.module.find('connectors'):
            name = connector.get('name')
            breadboard = connector.find('views').find('breadboardView')
            svgid = breadboard.find('p').get('svgId')

            anchorelm = self.imagexml.findall(f".//*[@id='{svgid}']")
            if len(anchorelm) == 0:
                anchorelm = self.imagexml.findall(f".//*[@id='{svgid}pin']")

            if len(anchorelm) == 0:
                return

            anchorelm = anchorelm[0]
            anchorx, anchory = self._anchor_position(name, anchorelm)

            if anchorx is not None and anchory is not None:
                self.anchors[name] = Point((anchorx, anchory))




''' Twoport elements made from groups of other elements '''
from __future__ import annotations
from functools import partial

from .compound import ElementCompound
from ..import elements as elm
from ..types import Point, Linestyle
from ..segments import Segment, SegmentText, Segment
from typing import Optional


tp_arrowlen = 1
tp_termlen = 0.5  # terminal length


class ElementTwoport(ElementCompound):
    ''' Compound twoport element

        Args:
            input_element: The element forming the input branch of the twoport
            output_element: The element forming the output branch of the twoport
            bpadx: Horizontal padding from edge of either component
            bpady: Vertical padding from edge of either component
            minw: Margin around component if smaller than minw
            terminals: Draw with terminals extending past box
            unit: Length of input and output element
            width: Width of the twoport box
            box: Draw twoport outline
            boxfill: Color to fill the twoport if not None
            boxlw: Line width of twoport outline
            boxls: Line style of twoport outline '-', '--', ':', etc.

        Anchors:
            * in_p
            * in_n
            * out_p
            * out_n
            * center
    '''
    def __init__(self, input_element: type[elm.Element2Term],
                 output_element: type[elm.Element2Term],
                 boxpadx: float = .2, boxpady: float = .2, minw: float = .5,
                 terminals: bool = True, unit: float = 1.5, width: float = 2.15,
                 box: bool = True, boxfill: Optional[str] = None, boxlw: Optional[float] = None, boxls: Optional[Linestyle] = None, **kwargs):
        self.input_element = input_element
        self.output_element = output_element
        self.boxpadx = boxpadx
        self.boxpady = boxpady
        self.minw = minw
        self.terminals = terminals
        self.unit = unit
        self.width = width
        self.box = box
        self.boxfill = boxfill
        self.boxlw = boxlw
        self.boxls = boxls
        super().__init__(unit=unit, **kwargs)

    def setup(self):
        self.input_component = self.input_element().down()
        self.output_component = self.output_element().down()

        bbox_input = self.input_component.get_bbox()
        bbox_output = self.output_component.get_bbox()

        # since components are not yet placed, transform is not taken into account for bbox calculation
        # hence, use the height rather than the width. Also assumes symmetry
        in_w = max((bbox_input.ymax - bbox_input.ymin)/2,  self.minw / 2)
        out_w = max((bbox_output.ymax - bbox_output.ymin)/2, self.minw / 2)

        self.add(self.input_component.at((self.boxpadx + in_w, 0)))
        self.add(self.output_component.at((self.width-self.boxpadx - out_w, 0)))

        # draw outline
        bbox = self.get_bbox()
        if self.box:
            self.add(elm.Rect(theta=0, at=[0, 0],
                              corner1=(0, bbox.ymin - self.boxpady), corner2=(self.width, bbox.ymax + self.boxpady),
                              fill=self.boxfill, lw=self.boxlw, ls=self.boxls, zorder=0))

        bbox = self.get_bbox()

        out_p = self.add(elm.Line().at(self.output_component.start).tox(bbox.xmax).right())
        out_n = self.add(elm.Line().at(self.output_component.end).tox(bbox.xmax).right())
        in_p = self.add(elm.Line().at(self.input_component.start).tox(bbox.xmin).left())
        in_n = self.add(elm.Line().at(self.input_component.end).tox(bbox.xmin).left())

        self.anchors['in_p'] = in_p.end
        self.anchors['in_n'] = in_n.end
        self.anchors['out_p'] = out_p.end
        self.anchors['out_n'] = out_n.end
        self.anchors['center'] = ((bbox.xmin + bbox.xmax) / 2, (bbox.ymin + bbox.ymax) / 2)

        if self.terminals:
            for anchor in ['in_p', 'in_n', 'out_p', 'out_n']:
                if 'in' in anchor:
                    xadjust = -tp_termlen
                else:
                    xadjust = tp_termlen

                previous_anchor = self.anchors[anchor]
                new_anchor = (previous_anchor[0] + xadjust, previous_anchor[1])
                self.anchors[anchor] = new_anchor

                self.segments.append(Segment([previous_anchor, new_anchor]))


class TwoPort(ElementTwoport):
    ''' Generic Twoport

        Args:
            bpadx: Horizontal padding from edge of either component
            bpady: Vertical padding from edge of either component
            minw: Margin around component if smaller than minw
            terminals: Draw with terminals extending past box
            component_offset: Offset between input and output element
            box: Draw twoport outline
            boxfill: Color to fill the twoport if not None
            boxlw: Line width of twoport outline
            boxls: Line style of twoport outline '-', '--', ':', etc.
            sign: Draw input and output terminal labels
            arrow: Draw arrow from input to output

        Anchors:
            * in_p
            * in_n
            * out_p
            * out_n
            * center
    '''
    def __init__(self, sign: bool = True, arrow: bool = True, reverse_output: bool = False, **kwargs):
        self.sign = sign
        self.arrow = arrow
        self.reverse_output = reverse_output
        super().__init__(input_element=elm.Gap, output_element=elm.Gap,
                         boxpadx=0, minw=0, component_offset=2, **kwargs)

    def setup(self):
        super().setup()
        if self.sign:
            # note the use of unicode − rather than usual - for better visual representation.
            self.segments.append(
                SegmentText(pos=Point((0.1, 0.1)) + self.input_component.end, label='−',
                            align=('left', 'center'), rotation_global=False))
            self.segments.append(
                SegmentText(pos=Point((0.1, -0.1)) + self.input_component.start, label='+',
                            align=('left', 'center'), rotation_global=False))

            if not self.reverse_output:
                self.segments.append(
                    SegmentText(pos=Point((-0.1, 0.1)) + self.output_component.end, label='−',
                                align=('right', 'center'), rotation_global=False))
                self.segments.append(
                    SegmentText(pos=Point((-0.1, -0.1)) + self.output_component.start, label='+',
                                align=('right', 'center'), rotation_global=False))
            else:
                self.segments.append(
                    SegmentText(pos=Point((-0.1, 0.1)) + self.output_component.end, label='+',
                                align=('right', 'center'), rotation_global=False))
                self.segments.append(
                    SegmentText(pos=Point((-0.1, -0.1)) + self.output_component.start, label='−',
                                align=('right', 'center'), rotation_global=False))

        if self.arrow:
            center_point = Point(self.anchors['center'])
            self.segments.append(
                Segment((center_point + Point((-tp_arrowlen / 2, 0)), center_point + Point((tp_arrowlen / 2, 0))),
                        arrow='->',
                        lw=2,
                        arrowwidth=0.3,
                        arrowlength=0.3,
                        ))


class VoltageTransactor(ElementTwoport):
    ''' Voltage transactor

        Args:
            bpadx: Horizontal padding from edge of either component
            bpady: Vertical padding from edge of either component
            minw: Margin around component if smaller than minw
            terminals: Draw with terminals extending past box
            component_offset: Offset between input and output element
            box: Draw twoport outline
            boxfill: Color to fill the twoport if not None
            boxlw: Line width of twoport outline
            boxls: Line style of twoport outline '-', '--', ':', etc.
            reverse_output: Switch direction of output source, defaults to False

        Anchors:
            * in_p
            * in_n
            * out_p
            * out_n
            * center
    '''
    def __init__(self, reverse_output: bool = False, **kwargs):
        output_element: type[elm.Element2Term] = elm.SourceControlledV
        if not reverse_output:
            # mypy doesn't like partials
            output_element = partial(output_element, reverse=True)  # type:ignore
        super().__init__(input_element=elm.Gap, output_element=output_element, **kwargs)

    def setup(self):
        super().setup()
        self.segments.append(
            SegmentText(pos=Point((0, 0.05)) + self.input_component.end, label='−',
                        align=('right', 'bottom'), rotation_global=False))
        self.segments.append(
            SegmentText(pos=Point((0, -0.05)) + self.input_component.start, label='+',
                        align=('right', 'top'), rotation_global=False))


class TransimpedanceTransactor(ElementTwoport):
    ''' Transimpedance transactor

        Args:
            bpadx: Horizontal padding from edge of either component
            bpady: Vertical padding from edge of either component
            minw: Margin around component if smaller than minw
            terminals: Draw with terminals extending past box
            component_offset: Offset between input and output element
            box: Draw twoport outline
            boxfill: Color to fill the twoport if not None
            boxlw: Line width of twoport outline
            boxls: Line style of twoport outline '-', '--', ':', etc.
            reverse_output: Switch direction of output source, defaults to False

        Anchors:
            * in_p
            * in_n
            * out_p
            * out_n
            * center
    '''
    def __init__(self, reverse_output: bool = False, **kwargs):
        output_element: type[elm.Element2Term] = elm.SourceControlledV
        if not reverse_output:
            # element is reversed in itself, so do a double reversal to cancel
            output_element = partial(output_element, reverse=True)   # type:ignore
        super().__init__(input_element=elm.Line, output_element=output_element, **kwargs)

    def setup(self):
        super().setup()
        current_label_inline = elm.CurrentLabelInline(ofst=-0.15, headlength=0.3)
        current_label_inline.at(self.input_component)
        self.add(current_label_inline)


class CurrentTransactor(ElementTwoport):
    ''' Current transactor

        Args:
            bpadx: Horizontal padding from edge of either component
            bpady: Vertical padding from edge of either component
            minw: Margin around component if smaller than minw
            terminals: Draw with terminals extending past box
            component_offset: Offset between input and output element
            box: Draw twoport outline
            boxfill: Color to fill the twoport if not None
            boxlw: Line width of twoport outline
            boxls: Line style of twoport outline '-', '--', ':', etc.
            reverse_output: Switch direction of output source, defaults to False

        Anchors:
            * in_p
            * in_n
            * out_p
            * out_n
            * center
    '''

    def __init__(self, reverse_output: bool = False, **kwargs):
        output_element: type[elm.Element2Term] = elm.SourceControlledI
        if not reverse_output:
            # element is reversed in itself, so do a double reversal to cancel
            output_element = partial(output_element, reverse=True)  # type:ignore
        super().__init__(input_element=elm.Line, output_element=output_element, **kwargs)

    def setup(self):
        super().setup()
        current_label_inline = elm.CurrentLabelInline(ofst=-0.15, headlength=0.3)
        current_label_inline.at(self.input_component)
        self.add(current_label_inline)


class TransadmittanceTransactor(ElementTwoport):
    ''' Transadmittance transactor

        Args:
            bpadx: Horizontal padding from edge of either component
            bpady: Vertical padding from edge of either component
            minw: Margin around component if smaller than minw
            terminals: Draw with terminals extending past box
            component_offset: Offset between input and output element
            box: Draw twoport outline
            boxfill: Color to fill the twoport if not None
            boxlw: Line width of twoport outline
            boxls: Line style of twoport outline '-', '--', ':', etc.
            reverse_output: Switch direction of output source, defaults to False

        Anchors:
            * in_p
            * in_n
            * out_p
            * out_n
            * center
    '''

    def __init__(self, reverse_output: bool = False, **kwargs):
        output_element: type[elm.Element2Term] = elm.SourceControlledI
        self.reverse_output = reverse_output
        if not reverse_output:
            # element is reversed in itself, so do a double reversal to cancel
            output_element = partial(output_element, reverse=True)  # type:ignore
        super().__init__(input_element=elm.Gap, output_element=output_element, **kwargs)

    def setup(self):
        super().setup()
        self.segments.append(
            SegmentText(pos=Point((0, 0.05)) + self.input_component.end, label='−',
                        align=('right', 'bottom'), rotation_global=False))
        self.segments.append(
            SegmentText(pos=Point((0, -0.05)) + self.input_component.start, label='+',
                        align=('right', 'top'), rotation_global=False))


class Nullor(ElementTwoport):
    ''' Nullor

        Args:
            bpadx: Horizontal padding from edge of either component
            bpady: Vertical padding from edge of either component
            minw: Margin around component if smaller than minw
            terminals: Draw with terminals extending past box
            component_offset: Offset between input and output element
            box: Draw twoport outline
            boxfill: Color to fill the twoport if not None
            boxlw: Line width of twoport outline
            boxls: Line style of twoport outline '-', '--', ':', etc.

        Anchors:
            * in_p
            * in_n
            * out_p
            * out_n
            * center
    '''
    def __init__(self, **kwargs):
        super().__init__(input_element=elm.Nullator, output_element=elm.Norator, boxpadx=0.3, **kwargs)


class VMCMPair(ElementTwoport):
    ''' Nullor

        Args:
            bpadx: Horizontal padding from edge of either component
            bpady: Vertical padding from edge of either component
            minw: Margin around component if smaller than minw
            terminals: Draw with terminals extending past box
            component_offset: Offset between input and output element
            box: Draw twoport outline
            boxfill: Color to fill the twoport if not None
            boxlw: Line width of twoport outline
            boxls: Line style of twoport outline '-', '--', ':', etc.

        Anchors:
            * in_p
            * in_n
            * out_p
            * out_n
            * center
    '''
    def __init__(self, **kwargs):
        super().__init__(input_element=elm.VoltageMirror, output_element=elm.CurrentMirror,
                         boxpadx=0.3, **kwargs)



''' Lines, Arrows, and Labels '''

from __future__ import annotations
from typing import Optional, Sequence, Union
import math

from ..segments import Segment, SegmentCircle, SegmentArc, SegmentPoly, SegmentBezier
from .elements import Element, Element2Term
from .twoterm import gap
from ..types import XY, Point, Arcdirection, BilateralDirection, Linestyle
from .. import util
from .. import drawing_stack


class Line(Element2Term):
    r''' Straight Line

        Args:
            arrow: arrowhead specifier, such as '->', '<-', '<->', '-o', or '\|->'

        Keyword Args:
            arrowwidth: Width of arrowhead [default: .15]
            arrowlength: Length of arrowhead [default:  0.25]
    '''
    _element_defaults = {
        'arrowwidth': 0.15,
        'arrowlength': 0.25}
    def __init__(self, *, arrow: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)
        arrowwidth: float = self.params['arrowwidth']
        arrowlength: float = self.params['arrowlength']
        self.segments.append(Segment([(0, 0)], arrow=arrow,
                                     arrowwidth=arrowwidth, arrowlength=arrowlength))


bus_stroke = 0.25
class DataBusLine(Element2Term):
    ''' Straight Line with bus indication stripe '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment(
            [(0, 0), gap, (-bus_stroke/2, bus_stroke), (bus_stroke/2, -bus_stroke),
             gap, (0, 0)]))


class Arrow(Line):
    r''' Arrow

        Args:
            double: Show arrowhead on both ends

        Keyword Args:
            arrow: arrowhead specifier, such as '->', '<-', '<->', '-o', or '\|->'
            headwidth: Width of arrow head [default: 0.15]
            headlength: Length of arrow head [default: 0.25]
    '''
    def __init__(self, *,
                 double: bool = False,
                 headwidth: Optional[float] = None,
                 headlength: Optional[float] = None,
                 **kwargs):
        if double:
            kwargs.setdefault('arrow', '<->')
        else:
            kwargs.setdefault('arrow', '->')
        super().__init__(**kwargs)

        # Explicitly define center so reverses work
        self.anchors['center'] = (0, 0)


class Gap(Element2Term):
    ''' Gap for labeling port voltages, for example. Draws nothing,
        but provides place to attach a label such as ('+', 'V', '-').

        Keyword Args:
            lblloc: Label location within the gap [center]
            lblalign: Label alignment [(center, center)]
            lblofst: Offset to label [0]
    '''
    _element_defaults = {
        'lblloc': 'center',
        'lblalign': ('center', 'center'),
        'lblofst': 0
    }
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), gap, (1, 0)], visible=False))


class Dot(Element):
    ''' Connection Dot

        Keyword Args:
            radius: Radius of dot [default: 0.075]
            open: Draw as an open circle [default: False]
    '''
    _element_defaults = {
        'radius': 0.075,
        'open': False}
    def __init__(self, *,
                 radius: Optional[float] = None,
                 open: Optional[bool] = None,
                 **kwargs):
        super().__init__(**kwargs)
        fill = 'bg' if self.params['open'] else True
        self.anchors['start'] = (0, 0)
        self.anchors['center'] = (0, 0)
        self.anchors['end'] = (0, 0)
        self.elmparams['drop'] = (0, 0)
        self.elmparams['theta'] = 0
        self.elmparams['zorder'] = 4
        self.elmparams['fill'] = fill
        self.segments.append(SegmentCircle((0, 0), self.params['radius']))


class Arrowhead(Element):
    ''' Arrowhead 
    
        Args:
            headwidth: width of arrow head [default: .15]
            headlength: length of arrow head [default: .25]
    '''
    _element_defaults = {
        'headwidth': 0.15,
        'headlength': 0.25,
        'lblofst': 0.25
    }
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        headwidth = self.params['headwidth']
        headlength = self.params['headlength'] 
        self.segments.append(Segment([
            (-headlength, 0), (0, 0)], arrowwidth=headwidth, arrowlength=headlength, arrow='->'))
        self.anchors['start'] = (0, 0)
        self.anchors['center'] = (0, 0)
        self.anchors['end'] = (0, 0)


class DotDotDot(Element):
    ''' Ellipsis element

        Keyword Args:
            radius: Radius of dots [default 0.075]
            open: Draw dots as open circles [default: False]

        "Ellipsis" is a reserved keyword in Python used for slicing,
        thus the name DotDotDot.
    '''
    _element_defaults = {
        'radius': 0.075,
        'open': False,
    }
    def __init__(self, *,
                 radius: Optional[float] = None,
                 open: Optional[bool] = None,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        r: float = self.params['radius']
        fill = 'bg' if self.params['open'] else True
        self.elmparams['fill'] = fill
        self.segments.append(SegmentCircle((.5, 0), r))
        self.segments.append(SegmentCircle((1, 0), r))
        self.segments.append(SegmentCircle((1.5, 0), r))
        self.elmparams['drop'] = (2, 0)


class Wire(Element):
    ''' Connect the .at() and .to() positions with lines depending on shape

        Args:
            shape: Determines shape of wire:
                `-`: straight line
                `|-`: right-angle line starting vertically
                `-|`: right-angle line starting horizontally
                'z': diagonal line with horizontal end segments
                'N': diagonal line with vertical end segments
                `n`: n- or u-shaped lines
                `c`: c- or ↄ-shaped lines
            k: Distance before the wire changes directions in `n` and `c` shapes.
            arrow: arrowhead specifier, such as '->', '<-', '<->', or '-o'
    '''
    def __init__(self, shape: str = '-', k: float = 1, *,
                 arrow: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)
        self._userparams['shape'] = shape
        self._userparams['k'] = k
        self._userparams.setdefault('to', (3, -2))

    def to(self, xy: XY, dx: float = 0, dy: float = 0) -> 'Element':
        ''' Specify ending position

            Args:
                xy: Ending position of element
                dx: X-offset from xy position
                dy: Y-offset from xy position
        '''
        xy = Point(xy)
        self._userparams['to'] = Point((xy.x + dx, xy.y + dy))
        return self

    def delta(self, dx: float = 0, dy: float = 0) -> 'Element':
        ''' Specify ending position relative to start position '''
        self._userparams['delta'] = Point((dx, dy))
        return self

    def dot(self, open: bool = False) -> 'Element':
        ''' Add a dot to the end of the element '''
        self._userparams['dot'] = True if not open else 'open'
        return self

    def idot(self, open: bool = False) -> 'Element':
        ''' Add a dot to the input/start of the element '''
        self._userparams['idot'] = True if not open else 'open'
        return self

    def _place(self, dwgxy: XY, dwgtheta: float, **dwgparams) -> tuple[Point, float]:
        ''' Calculate absolute placement of Element '''
        self._dwgparams.clear()
        self._dwgparams.update(dwgparams)
        if not self._positioned:
            self._position()

        self.params['theta'] = 0
        xy = self.params.get('at', dwgxy)
        to = self.params.get('to', dwgxy)
        delta = self.params.get('delta', None)
        arrow = self.params.get('arrow', None)
        shape = self.params.get('shape', '-')
        k = self.params.get('k', 1)
        if delta is not None:
            dx, dy = delta
        else:
            dx = to[0] - xy[0]
            dy = to[1] - xy[1]

        if shape == '-':     # Straight line
            self.segments.append(Segment([(0, 0), (dx, dy)], arrow=arrow))
            self.anchors['mid'] = (dx/2, dy/2)
        elif shape == '-|':  # Right angle, horizontal first
            self.segments.append(Segment([(0, 0), (dx, 0), (dx, dy)], arrow=arrow))
            self.anchors['mid'] = (dx, dy/2)
            self.elmparams['droptheta'] = 90 if dy > 0 else -90
        elif shape == '|-':  # Right angle, vertical first
            self.segments.append(Segment([(0, 0), (0, dy), (dx, dy)], arrow=arrow))
            self.anchors['mid'] = (dx/2, dy)
            self.elmparams['droptheta'] = 0 if dx > 0 else 180
        elif shape in ['z', 'Z']:
            if dx > 0:
                k = abs(k)
            else:
                k = -abs(k)
            self.segments.append(Segment([(0, 0), (k, 0), (dx-k, dy), (dx, dy)], arrow=arrow))
            self.anchors['mid'] = (dx/2, dy/2)
            self.elmparams['droptheta'] = 0 if dx > 0 else 180
        elif shape == 'N':
            if dy > 0:
                k = abs(k)
            else:
                k = -abs(k)
            self.segments.append(Segment([(0, 0), (0, k), (dx, dy-k), (dx, dy)], arrow=arrow))
            self.anchors['mid'] = (dx/2, dy/2)
            self.params['droptheta'] = 90 if dy > 0 else -90
        elif shape == 'n':   # N-shape
            self.segments.append(Segment([(0, 0), (0, k), (dx, k), (dx, dy)], arrow=arrow))
            self.anchors['mid'] = (dx/2, k)
            self.params['droptheta'] = 90 if dy > k else -90
        elif shape == 'c':   # C-shape
            self.segments.append(Segment([(0, 0), (k, 0), (k, dy), (dx, dy)], arrow=arrow))
            self.anchors['mid'] = (k, dy/2)
            self.params['droptheta'] = 0 if dx > k else 180
        else:
            raise ValueError(f'Undefined shape parameter `{shape}`.')

        if self.params.get('dot', False):
            fill: Union[bool, str] = 'bg' if self.params['dot'] == 'open' else True
            self.segments.append(SegmentCircle((dx, dy), radius=0.075, fill=fill, zorder=3))
        if self.params.get('idot', False):
            fill = 'bg' if self.params['idot'] == 'open' else True
            self.segments.append(SegmentCircle((0, 0), radius=0.075, fill=fill, zorder=3))

        self.elmparams['lblloc'] = 'mid'
        self.elmparams['drop'] = (dx, dy)
        self.anchors['start'] = Point((0, 0))
        self.anchors['end'] = Point((dx, dy))
        return super()._place(dwgxy, dwgtheta, **dwgparams)


class Arc2(Element):
    ''' Arc Element

        Use `at` and `to` methods to define endpoints.

        Arc2 is a quadratic Bezier curve with control point
        halfway between the endpoints, generating a
        symmetric 'C' curve.

        Args:
            k: Control point factor. Higher k means more curvature.
            arrow: arrowhead specifier, such as '->', '<-', '<->', or '-o'

        Anchors:
            start
            end
            ctrl
            mid
    '''
    def __init__(self, k: float = 0.5, arrow: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)
        self.k = k
        self.arrow = arrow

    def to(self, xy: XY, dx: float = 0, dy: float = 0) -> 'Element':
        ''' Specify ending position

            Args:
                xy: Ending position of element
                dx: X-offset from xy position
                dy: Y-offset from xy position
        '''
        xy = Point(xy)
        self._userparams['to'] = Point((xy.x + dx, xy.y + dy))
        return self

    def delta(self, dx: float = 0, dy: float = 0) -> 'Element':
        ''' Specify change in position '''
        self._userparams['delta'] = Point((dx, dy))
        return self

    def _place(self, dwgxy: XY, dwgtheta: float, **dwgparams) -> tuple[Point, float]:
        ''' Calculate absolute placement of Element '''
        self._dwgparams = dwgparams
        if not self._positioned:
            self._position()

        xy: Point = Point(self.params.get('at', dwgxy))
        to: Point = Point(self.params.get('to', Point((xy.x+3, xy.y))))
        delta = self.params.get('delta', None)
        if delta is not None:
            dx, dy = delta
        else:
            dx = to.x - xy.x
            dy = to.y - xy.y
        pa = Point((dx/2-dy*self.k, dy/2+dx*self.k))
        mid = Point((dx/2-dy*self.k/2, dy/2+dx*self.k/2))
        self.segments.append(SegmentBezier(
            (Point((0, 0)), pa, Point((dx, dy))), arrow=self.arrow))

        self.anchors['ctrl'] = pa
        self.anchors['mid'] = mid
        self.elmparams['theta'] = 0
        self.elmparams['lblloc'] = 'mid'
        self.anchors['start'] = Point((0, 0))
        self.anchors['end'] = Point((dx, dy))
        self.elmparams['drop'] = Point((dx, dy))
        valign = 'bottom'
        halign = 'left'
        vofst = 0.1
        if math.isclose(mid.y, pa.y, abs_tol=.01):
            valign = 'center'
            vofst = 0
        elif mid.y > pa.y:
            valign = 'top'
            vofst = -0.1
        if math.isclose(mid.x, pa.x, abs_tol=.01):
            halign = 'center'
        elif mid.x > pa.x:
            halign = 'right'
        self.elmparams['lblofst'] = vofst
        self.elmparams['lblalign'] = (halign, valign)
        return super()._place(dwgxy, dwgtheta, **dwgparams)


class Arc3(Element):
    ''' Arc Element

        Use `at` and `to` methods to define endpoints.

        Arc3 is a cubic Bezier curve. Control points are set
        to extend the curve at the given angle for each endpoint.

        Args:
            k: Control point factor. Higher k means tighter curve.
            th1: Angle at which the arc leaves start point
            th2: Angle at which the arc leaves end point
            arrow: arrowhead specifier, such as '->', '<-', '<->', or '-o'
            arrowlength: Length of arrowhead [default: 0.25]
            arrowwidth: Width of arrowhead [default: 0.2]

        Anchors:
            start
            end
            center
            ctrl1
            ctrl2
    '''
    _element_defaults = {
        'arrowlength': 0.25,
        'arrowwidth': 0.2
    }
    def __init__(self,
                 k: float = 0.75,
                 th1: float = 0.,
                 th2: float = 180,
                 arrow: Optional[str] = None,
                 *,
                 arrowlength: Optional[float] = None,
                 arrowwidth: Optional[float] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self._userparams.setdefault('to', Point((1, 0)))
        self.k = k
        self.th1 = math.radians(th1)
        self.th2 = math.radians(th2)
        self.arrow = arrow

    def to(self, xy: XY, dx: float = 0, dy: float = 0) -> 'Element':
        ''' Specify ending position

            Args:
                xy: Ending position of element
                dx: X-offset from xy position
                dy: Y-offset from xy position
        '''
        xy = Point(xy)
        self._userparams['to'] = Point((xy.x + dx, xy.y + dy))
        return self

    def delta(self, dx: float = 0, dy: float = 0) -> 'Element':
        ''' Specify change in position '''
        self._userparams['delta'] = Point((dx, dy))
        return self

    def _place(self, dwgxy: XY, dwgtheta: float, **dwgparams) -> tuple[Point, float]:
        ''' Calculate absolute placement of Element '''
        self._dwgparams = dwgparams
        if not self._positioned:
            self._position()

        self.params['theta'] = 0
        xy: Point = Point(self.params.get('at', dwgxy))
        to: Point = Point(self.params.get('to', dwgxy))
        delta = self.params.get('delta', None)
        if delta is not None:
            dx, dy = delta
        else:
            dx = to.x - xy.x
            dy = to.y - xy.y
        pa1 = Point((dx*self.k*math.cos(self.th1),
                     dy*self.k*math.sin(self.th1)))
        pa2 = Point((dx+dx*self.k*math.cos(self.th2),
                     dy+dy*self.k*math.sin(self.th2)))
        self.segments.append(SegmentBezier((Point((0, 0)), pa1, pa2, Point((dx, dy))),
                                           arrow=self.arrow,
                                           arrowlength=self.params['arrowlength'],
                                           arrowwidth=self.params['arrowwidth']))
        self.anchors['center'] = Point((dx/2, dy/2))
        self.anchors['start'] = Point((0, 0))
        self.anchors['end'] = Point((dx, dy))
        self.anchors['ctrl1'] = pa1
        self.anchors['ctrl2'] = pa2
        self.elmparams['drop'] = Point((dx, dy))
        self._setlabel(dx, dy)
        return super()._place(dwgxy, dwgtheta, **dwgparams)

    def _setlabel(self, dx: float, dy: float) -> None:
        ''' Set label position/alignment '''
        # This sets reasonable label position along center of arc.
        self.elmparams['lblloc'] = 'center'
        halign = 'left'
        valign = 'bottom'
        if dy > 0:
            valign = 'top'
            self.elmparams['lblofst'] = -0.1
        if dx <= 0:
            halign = 'right'
        self.elmparams['lblalign'] = (halign, valign)


class Annotate(Arc3):
    ''' Draw a curved arrow pointing to `at` position, ending at `to`
        position, with label location at the tail of the arrow
        (See also `Arc3`).

        Args:
            k: Control point factor. Higher k means tighter curve.
            th1: Angle at which the arc leaves start point
            th2: Angle at which the arc leaves end point
            arrow: arrowhead specifier, such as '->', '<-', '<->', or '-o'
            arrowlength: Length of arrowhead [default: 0.25]
            arrowwidth: Width of arrowhead [default: 0.2]
    '''
    def __init__(self,
                 k: float = 0.75,
                 th1: float = 75,
                 th2: float = 180,
                 arrow: str = '<-',
                 *,
                 arrowlength: Optional[float] = None,
                 arrowwidth: Optional[float] = None,
                 **kwargs):
        super().__init__(k=k, th1=th1, th2=th2, arrow=arrow,
                         arrowlength=arrowlength, arrowwidth=arrowwidth, **kwargs)
        self._userparams['to'] = Point((1, 1))

    def _setlabel(self, dx: float, dy: float) -> None:
        ''' Set label position/alignment '''
        # Attempt to align the label at the tail of the arrow.
        self.elmparams['lblloc'] = 'end'
        p1, p2 = Point(self.anchors['ctrl2']), Point(self.anchors['end'])
        if p1 == p2:
            th2 = -90. if dy < 0 else 90.
        else:
            th2 = math.degrees(math.atan2(p2.y-p1.y, p2.x-p1.x))
        while th2 < 0:
            th2 += 360.

        if th2 < 45 or th2 > 315:
            self.elmparams['lblalign'] = ('left', 'center')
            self.elmparams['lblofst'] = (.1, 0)
        elif 45 <= th2 <= 135:
            self.elmparams['lblalign'] = ('center', 'bottom')
            self.elmparams['lblofst'] = (0, .1)
        elif 135 < th2 <= 225:
            self.elmparams['lblalign'] = ('right', 'center')
            self.elmparams['lblofst'] = (-.1, 0)
        else:
            self.elmparams['lblalign'] = ('center', 'top')
            self.elmparams['lblofst'] = (0, -.1)


class ArcZ(Arc3):
    ''' Z-Curve Arc

        Use `at` and `to` methods to define endpoints.

        ArcZ approaches the endpoints horizontally, leading to
        a 'Z' shaped curve

        Args:
            k: Control point factor. Higher k means tighter curve.
            arrow: arrowhead specifier, such as '->', '<-', '<->', or '-o' [default: None]
            arrowlength: Length of arrowhead [default: 0.25]
            arrowwidth: Width of arrowhead [default: 0.2]
    '''
    def __init__(self,
                 k: float = 0.75,
                 arrow: Optional[str] = None,
                 *,
                 arrowlength: Optional[float] = None,
                 arrowwidth: Optional[float] = None,
                 **kwargs):
        super().__init__(k=k,
                         arrow=arrow,
                         th1=0,
                         th2=180,
                         arrowlength=arrowlength,
                         arrowwidth=arrowwidth,
                         **kwargs)


class ArcN(Arc3):
    ''' N-Curve Arc

        Use `at` and `to` methods to define endpoints.

        ArcN approaches the endpoints vertically, leading to
        a 'N' shaped curve

        Args:
            k: Control point factor. Higher k means tighter curve.
            arrow: arrowhead specifier, such as '->', '<-', '<->', or '-o' [default: None]
            arrowlength: Length of arrowhead [default: 0.25]
            arrowwidth: Width of arrowhead [default: 0.2]
    '''
    def __init__(self,
                 k: float = 0.75,
                 arrow: Optional[str] = None,
                 *,
                 arrowlength: Optional[float] = None,
                 arrowwidth: Optional[float] = None,
                 **kwargs):
        super().__init__(k=k,
                         arrow=arrow,
                         th1=90,
                         th2=-90,
                         arrowlength=arrowlength,
                         arrowwidth=arrowwidth,
                         **kwargs)


class ArcLoop(Element):
    ''' Loop Arc

        Use `at` and `to` methods to define endpoints.

        ArcLoop is an arc drawn as part of a circle.

        Args:
            radius: Radius of the arc
            arrow: arrowhead specifier, such as '->', '<-', '<->', or '-o' [default: None]
            arrowlength: Length of arrowhead [default: 0.25]
            arrowwidth: Width of arrowhead [default: 0.2]

        Anchors:
            start
            end
            mid
            BL
            BR
            TL
            TR
    '''
    _element_defaults = {
        'arrowlength': 0.25,
        'arrowwidth': 0.2
    }
    def __init__(self,
                 radius: float = 0.6,
                 arrow: Optional[str] = None,
                 *,
                 arrowlength: Optional[float] = None,
                 arrowwidth: Optional[float] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.radius = radius
        self.arrow = arrow

    def to(self, xy: XY, dx: float = 0, dy: float = 0) -> 'Element':
        ''' Specify ending position

            Args:
                xy: Ending position of element
                dx: X-offset from xy position
                dy: Y-offset from xy position
        '''
        xy = Point(xy)
        self._userparams['to'] = Point((xy.x + dx, xy.y + dy))
        return self

    def delta(self, dx: float = 0, dy: float = 0) -> 'Element':
        ''' Specify ending position relative to start position '''
        self._userparams['delta'] = Point((dx, dy))
        return self

    def _place(self, dwgxy: XY, dwgtheta: float, **dwgparams) -> tuple[Point, float]:
        ''' Calculate placement of Element '''
        self._dwgparams = dwgparams
        if not self._positioned:
            self._position()

        self.params['theta'] = 0
        xy: Point = Point(self.params.get('at', dwgxy))
        to: Point = Point(self.params.get('to', dwgxy))
        delta = self.params.get('delta', None)
        if delta is not None:
            dx, dy = delta
        else:
            dx = to.x - xy.x
            dy = to.y - xy.y

        xa = dx/2
        ya = dy/2
        q = math.sqrt(dx**2 + dy**2)
        try:
            center = Point(((xa + math.sqrt(self.radius**2 - (q/2)**2) * (-dy/q)),
                            (ya + math.sqrt(self.radius**2 - (q/2)**2) * (dx/q))))
        except ValueError as err:
            raise ValueError(f'No solution to ArcLoop with radius {self.radius}') from err

        da = math.sqrt(xa**2 + ya**2)
        thetaa = math.atan2(ya, xa)
        thda = math.acos(da/self.radius)
        theta1 = thetaa - thda
        theta2 = thetaa + thda
        t = util.linspace(math.pi+theta2, theta1)
        x = [center.x + self.radius * math.cos(ti) for ti in t]
        y = [center.y + self.radius * math.sin(ti) for ti in t]
        self.segments.append(Segment(list(zip(x, y)), arrow=self.arrow,
                                     arrowlength=self.params['arrowlength'],
                                     arrowwidth=self.params['arrowwidth']))
        mid = len(x)//2
        self.anchors['start'] = Point((0, 0))
        self.anchors['end'] = Point((x[-1], y[-1]))
        self.anchors['mid'] = Point((x[mid], y[mid]))
        self.anchors['BL'] = Point((min(x), min(y)))
        self.anchors['BR'] = Point((max(x), min(y)))
        self.anchors['TL'] = Point((min(x), max(y)))
        self.anchors['TR'] = Point((max(x), max(y)))
        self.elmparams['lblloc'] = 'mid'
        self.elmparams['drop'] = to
        valign = 'bottom'
        halign = 'left'
        if y[mid] < y[0]:
            valign = 'top'
            self.elmparams['lblofst'] = -.1
        if x[mid] < x[0] and x[mid] < x[-1]:
            halign = 'right'
        self.elmparams['lblalign'] = (halign, valign)
        return super()._place(dwgxy, dwgtheta, **dwgparams)


class Label(Element):
    ''' Label element.

        For more options, use `Label().label()` method.

        Args:
            label: text to display.
    '''
    def __init__(self, label: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)
        self.elmparams['lblloc'] = 'center'
        self.elmparams['lblofst'] = 0
        if label:
            self.label(label)


class Tag(Element):
    ''' Tag/flag element for labeling signal names.

        Because text size is unknown until drawn, must specify width
        manually to fit a given text label.

        Args:
            width: Width of the tag [default: 1.5]
            height: Height of the tag [default: 0.625]
    '''
    _element_defaults = {
        'width': 1.5,
        'height': 0.625,
        'lblloc': 'center',
        'fontsize': 12,
        'lblofst': 0
    }
    def __init__(self, *,
                 width: Optional[float] = None,
                 height: Optional[float] = 0.625,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        w: float = self.params['width']
        h: float = self.params['height']
        h = h / 2
        self.segments.append(SegmentPoly([(0, 0),
                                          (h, h),
                                          (w, h),
                                          (w, -h),
                                          (h, -h)]))
        self.anchors['start'] = (0, 0)


class CurrentLabel(Element):
    ''' Current label arrow drawn above an element

        Use `.at()` method to place the label over an
        existing element.

        Args:
            reverse: Reverse the arrow direction
            ofst: Offset distance from centerline of element [default: 0.4]
            length: Length of the arrow [default: 2]
            top: Draw arrow on top or bottom of element [default: True]
            headlength: Length of arrowhead [default: 0.3]
            headwidth: Width of arrowhead [default: 0.2]
    '''
    _element_defaults = {
        'ofst': 0.15,
        'length': 2,
        'top': True,
        'reverse': False,
        'headlength': 0.3,
        'headwidth': 0.2,
    }
    def __init__(self,
                 *,
                 length: Optional[float] = None,
                 top: Optional[bool] = None,
                 reverse: Optional[bool] = None,
                 ofst: Optional[float] = None,
                 headlength: Optional[float] = None,
                 headwidth: Optional[float] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.elmparams['lblofst'] = -.1
        self.elmparams['drop'] = None
        self.elmparams['anchor'] = 'center'
        self.anchors['center'] = (0, 0)
        self._side = 'top'

    def at(self, xy: XY | tuple['Element', str], dx: float = 0, dy: float = 0) -> 'Element':
        ''' Specify CurrentLabel position.

            If xy is an Element, arrow will be centered
            along element and its color will also be
            inherited.

            Args:
                xy: The absolute (x, y) position or an
                Element instance to center the arrow over
        '''
        if isinstance(xy, Element):
            side = xy.params.get('ilabel', 'top')
            if not self.params.get('top', True):
                side = {'top': 'bottom',
                        'bottom': 'top',
                        'left': 'right',
                        'right': 'left'}.get(side, 'top')                

            if xy.params.get('reverse', False):
                side = {'left': 'right',
                        'right': 'left'}.get(side, side)
                if side not in ['left', 'right']:
                    self.reverse()

            if xy.params.get('flip', False) and side in ['left', 'right']:
                self.reverse()

            if side == 'left':
                self.reverse()

            theta = xy.transform.theta
            bbox = xy.get_bbox(includetext=False)
            if side == 'top':
                pos = Point(((bbox.xmax + bbox.xmin)/2, bbox.ymax))
                self.elmparams['lblloc'] = 'top'
            elif side == 'bottom':
                pos = Point(((bbox.xmax + bbox.xmin)/2, bbox.ymin))
                self.elmparams['lblloc'] = 'bot'
            elif side == 'right':
                pos = Point(((bbox.xmax, (bbox.ymax + bbox.ymin)/2)))
                self.elmparams['lblloc'] = 'bot'
                theta -= 90
            elif side == 'left':
                pos = Point(((bbox.xmin, (bbox.ymax + bbox.ymin)/2)))
                self.elmparams['lblloc'] = 'top'
                theta += 90
            
            pos = xy.transform.transform(pos)
            self._side = side
            super().at(pos)

            self.theta(theta)
            if 'color' in xy._userparams:
                self.color(xy._userparams.get('color'))
        else:
            super().at(xy)
        return self

    def _place(self, dwgxy, dwgtheta, **dwgparams):
        ofst = self.params['ofst']
        length = self.params['length']
        if self._side in ['bottom']:
            ofst = -ofst

        theta = self.params.get('theta', 0) % 360
        loc = self.params.get('lblloc', 'top')

        if self._side in ['right'] and (theta <= 90 or theta > 270):
            loc = self.elmparams.get('lblloc', 'top')
            self.elmparams['lblloc'] = {'bot': 'top',
                                        'top': 'bot'}.get(loc)

        elif self._side in ['top', 'left', 'bottom'] and (90 < theta <= 270):
            loc = self.elmparams.get('lblloc', 'top')
            self.elmparams['lblloc'] = {'bot': 'top',
                                        'top': 'bot'}.get(loc)

        a, b = (-length/2, ofst), (length/2, ofst)
        headwidth = self.params['headwidth']
        headlength = self.params['headlength']
        self.segments.append(Segment((a, b), arrow='->', arrowwidth=headwidth, arrowlength=headlength))
        return super()._place(dwgxy, dwgtheta, **dwgparams)


class CurrentLabelInline(Element):
    ''' Current direction arrow, inline with element.

        Use `.at()` method to place arrow on an Element instance

        Args:
            direction: arrow direction 'in' or 'out' of element
            ofst: Offset along lead length
            start: Arrow at start or end of element
            headlength: Length of arrowhead [default: 0.3]
            headwidth: Width of arrowhead [default: 0.3]
    '''
    _element_defaults = {
        'headlength': 0.3,
        'headwidth': 0.3
    }
    def __init__(self,
                 direction: BilateralDirection = 'in',
                 ofst: float = 0.8,
                 start: bool = True,
                 *,
                 headlength: Optional[float] = None,
                 headwidth: Optional[float] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.elmparams['lblofst'] = 0
        self.elmparams['drop'] = None
        self.elmparams['zorder'] = 4

        x = ofst
        dx = self.params['headlength']
        if direction == 'in':
            x += self.params['headlength']
            dx = -dx

        if start:
            x = -x
            dx = -dx

        self.segments.append(Segment(((x, 0), (x+dx, 0)),
                                     arrow='->',
                                     arrowwidth=self.params['headwidth'],
                                     arrowlength=self.params['headlength']))

    def at(self, xy: XY | Element) -> 'Element':  # type: ignore[override]
        ''' Specify CurrentLabelInline position.

            If xy is an Element, arrow will be placed
            along the element's leads and the arrow color will
            be inherited.

            Args:
                xy: The absolute (x, y) position or an
                Element instance to place the arrow on
        '''
        if isinstance(xy, Element):
            super().at(xy.center)
            self.theta(xy.transform.theta)
            if 'color' in xy._userparams:
                self.color(xy._userparams.get('color'))  # type: ignore
        else:
            super().at(xy)
        return self


class ZLabel(Element):
    ''' Right-angle arrow, often used to indicate impedance
        looking in to a node

        Use `.at()` method to place the label over an
        existing element.

        Args:
            ofst: Vertical offset from centerline of element
            hofst: Horizontal offset from center of element
            length: Length of the arrow tail [default: 1]
            lengthtip: Length of the arrow tip [default: 0.5]
            headlength: Arrowhead length [default: 0.25]
            headwidth: Arrowhead width [default: 0.15]
    '''
    _element_defaults = {
        'ofst': 0.5,
        'hofst': 0.4,
        'length': 1,
        'lengthtip': 0.5,
        'headlength': 0.25,
        'headwidth': 0.15,
        'drop': None,
    }
    def __init__(self, *,
                 ofst: Optional[float] = None,
                 hofst: Optional[float] = None,
                 length: Optional[float] = None,
                 lengthtip: Optional[float] = None,
                 headlength: Optional[float] = None,
                 headwidth: Optional[float] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.anchor('center')
        self.anchors['center'] = (0, 0)

    def at(self, xy: XY | Element) -> 'Element':  # type: ignore[override]
        ''' Specify Element position.

            If xy is an Element, arrow will be centered
            along element and its color will also be
            inherited.

            Args:
                xy: The absolute (x, y) position or an
                Element instance to center the arrow over
        '''
        if isinstance(xy, Element):
            try:
                pos = xy.center
            except AttributeError:
                bbox = xy.get_bbox()
                pos = Point(((bbox.xmax + bbox.xmin)/2, (bbox.ymax + bbox.ymin)/2))

            super().at(pos)

            theta = xy.transform.theta
            if (theta % 360) > 90 and (theta % 360) <= 270:
                theta += 180  # Keeps 'top=True' labels above the element
            self.theta(theta)
            if 'color' in xy._userparams:
                self.color(xy._userparams.get('color'))  # type: ignore
        else:
            super().at(xy)
        return self

    def _place(self, dwgxy, dwgtheta, **dwgparams):
        self._dwgparams = dwgparams
        if not self._positioned:
            self._position()

        a = Point((-self.params['hofst'], self.params['ofst']))
        b = Point((-self.params['hofst']-self.params['lengthtip'], 
                   self.params['ofst']))
        c = Point((-self.params['hofst']-self.params['lengthtip'],
                   self.params['ofst']-self.params['length']))
        self.anchors['head'] = a
        self.anchors['tail'] = c

        self.segments.append(Segment((c, b, a), arrow='->',
                                     arrowwidth=self.params['headwidth'],
                                     arrowlength=self.params['headlength']))

        # Attempt to align the label at the tail of the arrow.
        self.elmparams['lblloc'] = 'tail'
        th = self.params.get('theta', 0)
        if self.params.get('flip'):
            th += 180

        th -= 90  # Because arrow is at right angle
        th = (th+360) % 360

        if th < 45 or th > 315:
            self.elmparams['lblalign'] = ('left', 'center')
            self.elmparams['lblofst'] = (0, -.1)
        elif 45 <= th <= 135:
            self.elmparams['lblalign'] = ('center', 'bottom')
            self.elmparams['lblofst'] = (0, .1)
        elif 135 < th <= 225:
            self.elmparams['lblalign'] = ('right', 'center')
            self.elmparams['lblofst'] = (0, .1)
        else:
            self.elmparams['lblalign'] = ('center', 'top')
            self.elmparams['lblofst'] = (0, -.1)

        return super()._place(dwgxy, dwgtheta, **dwgparams)


class LoopArrow(Element):
    ''' Loop arrow, for mesh analysis notation

        Args:
            direction: loop direction 'cw' or 'ccw'
            theta1: Angle of start of loop arrow
            theta2: Angle of end of loop arrow
            width: Width of loop
            height: Height of loop
    '''
    _element_defaults = {
        'lblloc': 'center',
        'lblofst': 0,
        'theta': 0,
    }
    def __init__(self, direction: Arcdirection = 'cw',
                 theta1: float = 35, theta2: float = -35,
                 width: float = 1.0, height: float = 1.0, **kwargs):
        super().__init__(**kwargs)

        self.segments.append(SegmentArc(
            (0, 0), arrow=direction, theta1=theta1, theta2=theta2,
            width=width, height=height))

        self.anchors['center'] = (0, 0)


class LoopCurrent(LoopArrow):
    ''' Loop current label, for mesh analysis notation,
        placed within a box of 4 existing elements.

        Args:
            elm_list: List of 4 elements surrounding loop, in
                      order (top, right, bottom, left)
            direction: loop direction 'cw' or 'ccw'
            theta1: Angle of start of loop arrow
            theta2: Angle of end of loop arrow
            pad: Distance from elements to loop
    '''
    def __init__(self, elm_list: Optional[Sequence[Element]] = None,
                 direction: Arcdirection = 'cw',
                 theta1: float = 35, theta2: float = -35,
                 pad: float = 0.2, **kwargs):
        assert elm_list is not None
        drawing_stack.push_element(None)  # Flush the stack to make sure all elm_list are placed
        bbox1 = elm_list[0].get_bbox(transform=True, includetext=False)
        bbox2 = elm_list[1].get_bbox(transform=True, includetext=False)
        bbox3 = elm_list[2].get_bbox(transform=True, includetext=False)
        bbox4 = elm_list[3].get_bbox(transform=True, includetext=False)
        top = bbox1.ymin - pad
        bot = bbox3.ymax + pad
        left = bbox4.xmax + pad
        rght = bbox2.xmin - pad
        center = ((left+rght)/2, (top+bot)/2)
        width = rght - left
        height = top - bot
        super().__init__(direction, theta1, theta2, width, height, **kwargs)
        self._userparams['at'] = center


class Rect(Element):
    ''' Rectangle Element

        Used mainly for buliding more complex elements. Corner
        arguments are relative to Element coordinates, not Drawing
        coordinates.

        Args:
            corner1: Position of top-left corner
            corner2: Position of bottom-right corner
            fill: Color to fill if not None [default: inherit]
            lw: Line width [default: inherit]
            ls: Line style '-', '--', ':', etc. [default: inherit]
    '''
    _element_defaults = {
        'zorder': 0,  # Put on bottom
    }
    def __init__(self,
                 corner1: XY = (0, 0),
                 corner2: XY = (1, 1),
                 *,
                 fill: Optional[str] = None,
                 lw: Optional[float] = None,
                 ls: Optional[Linestyle] = None,
                 **kwargs):
        super().__init__(**kwargs)
        c1a = (corner1[0], corner2[1])
        c2a = (corner2[0], corner1[1])
        self.segments.append(Segment([corner1, c1a, corner2, c2a, corner1], zorder=0))


class Encircle(Element):
    ''' Draw ellipse around all elements in the list

        Args:
            elm_list: List of elements to enclose
            padx: Horizontal distance from elements to loop [default: .2]
            pady: Vertical distance from elements to loop [default: .2]
            includelabels: Include labels in the ellipse
    '''
    _element_defaults = {
        'padx': 0.2,
        'pady': 0.2,
        'theta': 0
    }
    def __init__(self,
                 elm_list: Optional[Sequence[Element]] = None,
                 *,
                 padx: Optional[float] = None,
                 pady: Optional[float] = None,
                 includelabels: bool = True,
                 **kwargs):
        drawing_stack.push_element(None)  # Flush the stack to make sure all elm_list are placed
        super().__init__(**kwargs)
        assert elm_list is not None
        xmin = math.inf
        xmax = -math.inf
        ymin = math.inf
        ymax = -math.inf
        for element in elm_list:
            bbox = element.get_bbox(transform=True, includetext=includelabels)
            xmin = min(xmin, bbox.xmin)
            xmax = max(xmax, bbox.xmax)
            ymin = min(ymin, bbox.ymin)
            ymax = max(ymax, bbox.ymax)
        xmin -= self.params['padx']
        xmax += self.params['padx']
        ymin -= self.params['pady']
        ymax += self.params['pady']
        center = (xmax+xmin)/2, (ymax+ymin)/2
        w = xmax-xmin
        h = ymax-ymin
        self._userparams['at'] = center
        self.segments.append(SegmentArc((0, 0), w, h, theta1=0, theta2=360))
        sinpi4 = math.sin(math.pi/4)
        cospi4 = math.cos(math.pi/4)
        sinpi8 = math.sin(math.pi/8)
        cospi8 = math.cos(math.pi/8)
        self.anchors['N'] = (0, h/2)
        self.anchors['S'] = (0, -h/2)
        self.anchors['E'] = (w/2, 0)
        self.anchors['W'] = (-w/2, 0)
        self.anchors['SE'] = (w/2*cospi4, -h/2*sinpi4)
        self.anchors['SW'] = (-w/2*cospi4, -h/2*sinpi4)
        self.anchors['NW'] = (-w/2*cospi4, h/2*sinpi4)
        self.anchors['NE'] = (w/2*cospi4, h/2*sinpi4)
        self.anchors['ENE'] = (w/2*cospi8, h/2*sinpi8)
        self.anchors['WNW'] = (-w/2*cospi8, h/2*sinpi8)
        self.anchors['ESE'] = (w/2*cospi8, -h/2*sinpi8)
        self.anchors['WSW'] = (-w/2*cospi8, -h/2*sinpi8)
        self.anchors['NNE'] = (w/2*sinpi8, h/2*cospi8)
        self.anchors['NNW'] = (-w/2*sinpi8, h/2*cospi8)
        self.anchors['SSE'] = (w/2*sinpi8, -h/2*cospi8)
        self.anchors['SSW'] = (-w/2*sinpi8, -h/2*cospi8)


class EncircleBox(Element):
    ''' Draw rounded box around all elements in the list

        Args:
            elm_list: List elements to enclose
            cornerraidus: radius of corner rounding [default: 0.3]
            padx: Horizontal distance from elements to loop [default: 0.2]
            pady: Vertical distance from elements to loop [default: 0.2]
            includelabels: Include labels in the box
    '''
    _element_defaults = {
        'cornerradius': 0.3,
        'padx': 0.2,
        'pady': 0.2,
        'theta': 0
    }
    def __init__(self,
                 elm_list: Optional[Sequence[Element]] = None,
                 *,
                 cornerradius: float = 0.3,
                 padx: Optional[float] = None,
                 pady: Optional[float] = None,
                 includelabels: bool = True,
                 **kwargs):
        drawing_stack.push_element(None)  # Flush the stack to make sure all elm_list are placed
        super().__init__(**kwargs)
        assert elm_list is not None
        xmin = math.inf
        xmax = -math.inf
        ymin = math.inf
        ymax = -math.inf
        for element in elm_list:
            bbox = element.get_bbox(transform=True, includetext=includelabels)
            xmin = min(xmin, bbox.xmin)
            xmax = max(xmax, bbox.xmax)
            ymin = min(ymin, bbox.ymin)
            ymax = max(ymax, bbox.ymax)
        xmin -= self.params['padx']
        xmax += self.params['padx']
        ymin -= self.params['pady']
        ymax += self.params['pady']
        center = (xmax+xmin)/2, (ymax+ymin)/2
        w = xmax-xmin
        h = ymax-ymin
        self._userparams['at'] = center
        self.segments = [SegmentPoly(
            [(-w/2, h/2), (w/2, h/2), (w/2, -h/2), (-w/2, -h/2)],
            cornerradius=self.params['cornerradius'])]
        k = self.params['cornerradius'] - self.params['cornerradius']*math.sqrt(2)/2
        self.anchors['NE'] = (w/2-k, h/2-k)
        self.anchors['NW'] = (-w/2+k, h/2-k)
        self.anchors['SE'] = (w/2-k, -h/2+k)
        self.anchors['SW'] = (-w/2+k, -h/2+k)
        self.anchors['N'] = (0, h/2)
        self.anchors['S'] = (0, -h/2)
        self.anchors['E'] = (w/2, 0)
        self.anchors['W'] = (-w/2, 0)



''' Schemdraw base Element class '''

from __future__ import annotations
from typing import Sequence, MutableMapping, Any, Union, Optional
from collections import ChainMap
from dataclasses import dataclass
import warnings
import math

from .. import default_canvas
from ..segments import Segment, SegmentText, SegmentCircle, BBox, SegmentType
from ..transform import Transform
from .. import util
from ..util import Point
from ..types import XY, Linestyle, Halign, Valign, LabelLoc
from .. import drawing_stack

from ..backends.svg import Figure as svgFigure
try:
    from ..backends.mpl import Figure as mplFigure
except ImportError:
    mplFigure = None  # type: ignore


gap = (math.nan, math.nan)  # Put a gap in a path


@dataclass
class Label:
    ''' Element Label Parameters. '''
    label: str | Sequence[str]
    loc: LabelLoc | None = None   # top, bot, lft, rgt, OR anchor
    ofst: XY | float | None = None
    halign: Halign | None = None
    valign: Valign | None = None
    rotate: bool | float = False  # True=same angle as element; False = 0
    fontsize: float | None = None
    font: str | None = None
    mathfont: str | None = None
    color: str | None = None


class Element:
    ''' Standard circuit element.

        Keyword Arguments are equivalent to calling
        setter methods.

        Args:
            d: Drawing direction ('up', 'down', 'left', 'right')

        Attributes:
            anchors: Dictionary of anchor positions in element
                coordinates
            absanchors: Dictionary of anchor positions in absolute
                drawing coordinates
            segments: List of drawing primitives making up the element
            transform: Transformation from element to drawing coordinates
            absdrop: Drop position in drawing coordinates, set after the
                element is added to a drawing
            defaults: Default parameters for the element

        Anchor names are dynmically added as attributes after placing the
        element in a Drawing.
    '''
    _element_defaults: dict[str, Any] = {}     # Default parameters for subclassed elements
    defaults: ChainMap[str, Any] = ChainMap()  # Subclasses will chainmap this with parents  
    def __init__(self, **kwargs) -> None:
        self._userparams.update(kwargs)         # Specified by user
        self._localshift: XY = Point((0, 0))
        self._userlabels: list[Label] = []

        self.anchors: MutableMapping[str, Union[Point, tuple[float, float]]] = {}  # Untransformed anchors
        self.absanchors: MutableMapping[str, Any] = {}  # Transformed, absolute anchors
        self.segments: list[SegmentType] = []
        self.transform = Transform(0, (0, 0))
        self._positioned = False  # Has the element been placed in a drawing via self._position()?

        if 'xy' in self._userparams:  # Allow legacy 'xy' parameter
            self._userparams.setdefault('at', self._userparams.pop('xy'))

        drawing_stack.push_element(self)

    def __new__(cls, *args, **kwargs):
        ''' Create a new Element instance, building chainmap of params '''
        new = super().__new__(cls)
        new._dwgparams = {}  # Defaults from drawing
        new.elmparams = {}  # Parameters specified by element. Similar to _element_defaults, but may be dynamic
        new._userparams = {name: value for name, value in kwargs.items() if value is not None}
        new.params = ChainMap(new._userparams, new.elmparams, new.defaults, new._dwgparams)
        return new

    def __init_subclass__(self):
        ''' Initialize an Element subclass, building chainmap of default parameters
            from parent classes.
        '''
        if len(self.__mro__) > 1 and hasattr(self.__mro__[1], 'defaults'):
            if self.__mro__[1]._element_defaults is not self._element_defaults:
                self.defaults = self.__mro__[1].defaults.new_child(self._element_defaults)
            else:
                self.defaults = self.__mro__[1].defaults.new_child()
        else:
            self.defaults = ChainMap()

    def __getattr__(self, name: str) -> Any:
        ''' Allow getting anchor position as attribute '''
        anchornames = ['start', 'end', 'center', 'istart', 'iend',
                       'N', 'S', 'E', 'W', 'NE', 'NW', 'SE', 'SW',
                       'NNE', 'NNW', 'ENE', 'WNW', 'SSE', 'SSW', 'ESE', 'WSW']
        if (name in anchornames + list(vars(self).get('anchors', {}).keys()) and
            not name in vars(self).get('absanchors', {})):
                # Not placed yet
                drawing_stack.push_element(self)

        if name in vars(self).get('absanchors', {}):
            return vars(self).get('absanchors')[name]  # type: ignore
        raise AttributeError(f'{name} not defined in Element')

    def up(self) -> 'Element':
        ''' Set the direction to up '''
        if 'd' in self._userparams:
            warnings.warn(f"Duplicated direction parameter in element. `{self._userparams['d']}` changed to `up`.")
        self._userparams['d'] = 'up'
        return self

    def down(self) -> 'Element':
        ''' Set the direction to down '''
        if 'd' in self._userparams:
            warnings.warn("Duplicated direction parameter in element."
                          f" `{self._userparams['d']}` changed to `down`.")
        self._userparams['d'] = 'down'
        return self

    def left(self) -> 'Element':
        ''' Set the direction to left '''
        if 'd' in self._userparams:
            warnings.warn("Duplicated direction parameter in element."
                          f" `{self._userparams['d']}` changed to `left`.")
        self._userparams['d'] = 'left'
        return self

    def right(self) -> 'Element':
        ''' Set the direction to right '''
        if 'd' in self._userparams:
            warnings.warn("Duplicated direction parameter in element."
                          f" `{self._userparams['d']}` changed to `right`.")
        self._userparams['d'] = 'right'
        return self

    def theta(self, theta: float) -> 'Element':
        ''' Set the drawing direction angle in degrees '''
        if 'd' in self._userparams:
            warnings.warn("Duplicate direciton parameter in element:"
                          f" `{self._userparams['d']}` replaced with `theta={theta}`")
        self._userparams['theta'] = theta
        return self

    def drop(self, drop: Union[str, Point]) -> 'Element':
        ''' Set the drop position - where to leave the current drawing position after
            placing this element
        '''
        self._userparams['drop'] = drop
        return self

    def at(self, xy: XY | tuple['Element', str], dx: float = 0, dy: float = 0) -> 'Element':
        ''' Set the element xy position

            Args:
                xy: (x,y) position or tuple of (Element, anchorname)
        '''
        if 'at' in self._userparams:
            warnings.warn("Duplicate `at` parameter in element: "
                          f"`{self._userparams['at']}` changed to `{xy}`.")
        if isinstance(xy[1], str):
            self._userparams['at'] = xy
            if dx != 0 or dy != 0:
                raise ValueError('dx and dy must be zero for anchorname XY')
        else:
            xy = Point(xy)
            self._userparams['at'] = Point((xy.x + dx, xy.y + dy))
        return self

    def scale(self, scale: float = 1) -> 'Element':
        ''' Apply scale/zoom factor to element '''
        self._userparams['zoom'] = Point((scale, scale))
        return self

    def scalex(self, scale: float = 1) -> 'Element':
        ''' Apply horizontal scale/zoom to element '''
        oldscale = self._userparams.get('zoom', Point((1, 1)))
        self._userparams['zoom'] = Point((scale, oldscale.y))
        return self

    def scaley(self, scale: float = 1) -> 'Element':
        ''' Apply vertical scale/zoom to element '''
        oldscale = self._userparams.get('zoom', Point((1, 1)))
        self._userparams['zoom'] = Point((oldscale.x, scale))
        return self

    def flip(self) -> 'Element':
        ''' Apply flip up/down '''
        self._userparams['flip'] = True
        return self

    def reverse(self) -> 'Element':
        ''' Apply reverse left/right '''
        if 'reverse' in self._userparams:
            self._userparams['reverse'] = not self._userparams['reverse']
        else:
            self._userparams['reverse'] = True
        return self

    def anchor(self, anchor: str) -> 'Element':
        ''' Specify anchor for placement. The anchor will be
            aligned with the position specified by `at()` method.
        '''
        if 'anchor' in self._userparams:
            warnings.warn("Duplicate anchor parameter in element: "
                          f"`{self._userparams['anchor']}` changed to `{anchor}`.")

        self._userparams['anchor'] = anchor
        return self

    def color(self, color: str) -> 'Element':
        ''' Sets the element color

            Args:
                color: color name or hex value (ie '#FFFFFF')
        '''
        self._userparams['color'] = color
        return self

    def linestyle(self, ls: Linestyle) -> 'Element':
        ''' Sets the element line style

            Args:
                ls: Line style ('-', ':', '--', '-.').
        '''
        self._userparams['ls'] = ls
        return self

    def linewidth(self, lw: float) -> 'Element':
        ''' Sets the element line width

            Args:
                lw: Line width
        '''
        self._userparams['lw'] = lw
        return self

    def fill(self, color: bool | str = True) -> 'Element':
        ''' Sets the element fill color.

            Args:
                color: Color string name or hex value, or
                `True` to fill with the element line color.
        '''
        self._userparams['fill'] = color
        return self

    def style(self, color: Optional[str] = None, fill: Optional[str] = None,
              ls: Optional[Linestyle] = None, lw: Optional[float] = None) -> 'Element':
        ''' Apply all style parameters

            Args:
                color: Color string or hex value
                fill: Color string or hex
                ls: Line style ('-', ':', '--', '-.')
                lw: Line width
        '''
        if color is not None:
            self.color(color)
        if fill is not None:
            self.fill(fill)
        if ls is not None:
            self.linestyle(ls)
        if lw is not None:
            self.linewidth(lw)
        return self

    def zorder(self, zorder: int) -> 'Element':
        ''' Sets the element zorder. Higher zorders will be drawn above
            lower zorder elements.
        '''
        self._userparams['zorder'] = zorder
        return self

    def hold(self) -> 'Element':
        ''' Do not move the Drawing `here` position after placing this element '''
        self._userparams['move_cur'] = False
        return self

    def label(self,
              label: str | Sequence[str],
              loc: Optional[LabelLoc] = None,
              ofst: XY | float | None = None,
              halign: Optional[Halign] = None,
              valign: Optional[Valign] = None,
              rotate: bool | float = False,
              fontsize: Optional[float] = None,
              font: Optional[str] = None,
              mathfont: Optional[str] = None,
              color: Optional[str] = None):
        ''' Add a label to the Element.

            Args:
                label: The text string or list of strings. If list, each string will
                    be evenly spaced along the element (e.g. ['-', 'V', '+'])
                loc: Label position within the Element. Either ('top', 'bottom', 'left',
                    'right'), or the name of an anchor within the Element.
                ofst: Offset from default label position
                halign: Horizontal text alignment ('center', 'left', 'right')
                valign: Vertical text alignment ('center', 'top', 'bottom')
                rotate: True to rotate label with element, or specify rotation
                    angle in degrees
                fontsize: Size of label font
                font: Name/font-family of label text
                mathfont: Name/font-family of math text
                color: Color of label
        '''
        if not rotate:
            rotate = 0
        elif isinstance(rotate, bool):
            rotate = True
        self._userlabels.append(Label(label, loc, ofst, halign, valign, rotate, fontsize, font, mathfont, color))
        return self

    def _position(self) -> None:
        ''' Convert relative positions to absolute coordinates in self.params,
            and apply flip/reverse
        '''
        # Accomodate xy positions based on OTHER elements before they are fully set up.
        if 'at' in self._userparams and isinstance(self._userparams['at'][1], str):
            element, pos = self._userparams['at']
            if pos in element.absanchors:
                xy = element.absanchors[pos]
            else:
                raise KeyError(f'Unknown anchor name {pos}')
            self._userparams['at'] = xy

        self._flipreverse()
        self._positioned = True

    def _flipreverse(self) -> None:
        ''' Flip and/or reverse element's segments if necessary '''
        if self._userparams.get('flip', False):
            for s in self.segments:
                s.doflip()
            for name, pt in self.anchors.items():
                self.anchors[name] = Point(pt).flip()

        if self._userparams.get('reverse', False):
            if 'center' in self.anchors:
                centerx = self.anchors['center'][0]
            else:
                xmin, _, xmax, _ = self.get_bbox(includetext=False)
                centerx = (xmin + xmax)/2
            [s.doreverse(centerx) for s in self.segments]  # type: ignore
            for name, pt in self.anchors.items():
                self.anchors[name] = Point(pt).mirrorx(centerx)

    def _place(self, dwgxy: XY, dwgtheta: float, **dwgparams) -> tuple[Point, float]:
        ''' Calculate element position within the drawing
        
            Args:
                dwgxy: Current XY position within drawing
                dwgtheta: Current theta in the drawing
                dwgparams: Default parameters of the drawing
            
            Returns:
                xy: New XY position after placing the element
                theta: New theta after placing the element
        '''
        self._dwgparams.clear()  # Don't remove the original object so self.params ChainMap gets the new values.
        self._dwgparams.update(dwgparams)
        if not self._positioned:
            self._position()

        anchor = self.params.get('anchor', None)
        zoom = self.params.get('zoom', self.params.get('scale', 1))
        xy = self.params.get('at', dwgxy)

        # Get bounds of element, used for positioning user labels
        self.bbox = self.get_bbox(includetext=False)

        theta: float
        if 'endpts' in self.params:
            theta = dwgtheta
        elif self.params.get('d') is not None:
            d = self.params.get('d')
            if str(d).lstrip('-').isnumeric():
                theta = float(str(d))
            else:
                theta = {'u': 90, 'r': 0, 'l': 180, 'd': 270}[d[0].lower()]  # type: ignore
        else:
            theta = self.params.get('theta', dwgtheta)

        if anchor is not None:
            self._localshift = -Point(self.anchors[anchor])
        self.transform = Transform(theta, xy, self._localshift, zoom)

        # Add user-defined labels
        lblloc = self.params.get('lblloc', 'top')
        lblsize = self.params.get('lblsize', self.params.get('fontsize', 14))
        lblrotate = self.params.get('lblrotate', False)
        lblcolor = self.params.get('color', 'black')
        kwlabels = {
            'top': self.params.get('toplabel', None),
            'bot': self.params.get('botlabel', None),
            'lft': self.params.get('lftlabel', None),
            'rgt': self.params.get('rgtlabel', None),
            'center': self.params.get('clabel', None)
            }
        if 'label' in self.params:
            kwlabels[lblloc] = self.params.get('label')

        # Add labels defined in **kwargs to the _userlabels list
        for loc, label in kwlabels.items():
            if label is not None:
                rotate = (theta if lblrotate else 0)
                self.label(label, loc, fontsize=lblsize, rotate=rotate, color=lblcolor)

        for label in self._userlabels:
            self._place_label(label, theta)

        # Add element-specific anchors
        for name, pos in self.anchors.items():
            self.absanchors[name] = self.transform.transform(pos)
        self.absanchors['xy'] = self.transform.transform((0, 0))

        drop = self.params.get('drop', None)
        if drop is not None and drop in self.anchors:
            # User specified as anchor position
            self.absdrop = self.transform.transform(self.anchors[drop]), theta
        elif drop is not None and self.params.get('move_cur', True):
            if self.params.get('droptheta', None) is not None:
                # Element-specified drop angle
                self.absdrop = self.transform.transform(drop), self.params.get('droptheta', 0.)
            elif self.params.get('theta', None) == 0:
                # Element-specified theta of 0, don't change theta
                self.absdrop = self.transform.transform(drop), dwgtheta
            else:
                self.absdrop = self.transform.transform(drop), theta
        else:  # drop is None:
            self.absdrop = Point(dwgxy), dwgtheta
        return self.absdrop

    def get_bbox(self, transform=False, includetext=True):
        ''' Get element bounding box

            Args:
                transform: Apply the element transform to the bbox
                    to get bounds in Drawing coordinates
                includetext: Consider text when calculating bounding
                    box. Text width and height can vary by font,
                    so this produces an estimate of bounds.

            Returns:
                Corners of the bounding box, (xmin, ymin, xmax, ymax)
        '''
        xmin = ymin = math.inf
        xmax = ymax = -math.inf
        for segment in self.segments:
            if not includetext and isinstance(segment, SegmentText):
                continue
            if transform:
                segment = segment.xform(self.transform)
            segxmin, segymin, segxmax, segymax = segment.get_bbox()
            xmin = min(xmin, segxmin)
            xmax = max(xmax, segxmax)
            ymin = min(ymin, segymin)
            ymax = max(ymax, segymax)

        return BBox(xmin, ymin, xmax, ymax)

    def _position_label(self, label: Label, theta: float = 0) -> Label:
        ''' Calculate position of label
        
            Args:
                label: The label to position
                theta: Element drawing direction
        '''
        if label.rotate is None:
            label.rotate = 0
        elif label.rotate is True:
            label.rotate = theta
        label.rotate = (label.rotate + 360) % 360
        if 90 < label.rotate < 270:
            label.rotate -= 180  # Keep the label from going upside down

        # Set label default location if not specified
        if label.loc is None:
            label.loc = self.params.get('lblloc', 'top')

        # Allow some aliases for positional locations
        label.loc = {'bot': 'bottom',
                     'B': 'bottom',
                     'lft': 'left',
                     'L': 'left',
                     'rgt': 'right',
                     'R': 'right',
                     'T': 'top'}.get(label.loc, label.loc)

        # Ensure a 'top' label is always on top, regardless of rotation
        if (theta % 360) > 90 and (theta % 360) <= 270:
            if label.loc == 'top':
                label.loc = 'bottom'
            elif label.loc == 'bottom':
                label.loc = 'top'
            elif label.loc == 'left':
                label.loc = 'right'
            elif label.loc == 'right':
                label.loc = 'left'
        return label

    def _align_label(self, label: Label, theta: float = 0) -> tuple[Halign, Valign, Point]:
        ''' Calculate label alignment and offset based on angle and location
            relative to the element

            Args:
                label: The label to position
                theta: Element drawing direction

            Returns:
                Align: suggested horizontal and vertical alignment
                Offset: suggested horizontal and vertical offset from label position
        '''
        newofst = label.ofst
        assert newofst is not None
        if isinstance(label.ofst, (float, int)):
            newofst = Point((label.ofst, label.ofst))
        else:
            newofst = Point(newofst)  # type: ignore

        if label.loc == 'center':
            newhalign: Halign = 'center'
            newvalign: Valign = 'center'

        elif label.loc and label.loc in self.anchors: 
            # Anchor is on an edge
            x1, y1, x2, y2 = self.get_bbox(includetext=False)
            newhalign = newvalign = 'center'
            if math.isclose(self.anchors[label.loc][0], x1, abs_tol=.15):
                # Label on left edge
                newhalign = 'right'
            elif math.isclose(self.anchors[label.loc][0], x2, abs_tol=.15):
                # Label on right edge
                newhalign = 'left'
            else:
                # Not on left or right edge
                newhalign = 'center'

            if math.isclose(self.anchors[label.loc][1], y1, abs_tol=.15):
                # Label on bottom edge
                newvalign = 'top'
            elif math.isclose(self.anchors[label.loc][1], y2, abs_tol=.15):
                # Label on top edge
                newvalign = 'bottom'
            else:
                # Not on top or bottom edge
                newvalign = 'center'

            # Fix offset if provided as single value
            if isinstance(label.ofst, (float, int)):
                pofst = label.ofst
                newofst = {
                    ('center', 'bottom'): (0, pofst),
                    ('right', 'bottom'): (-pofst, pofst),
                    ('right', 'center'): (-pofst, 0),
                    ('right', 'top'): (-pofst, -pofst),
                    ('center', 'top'): (0, -pofst),
                    ('left', 'top'): (pofst, -pofst),
                    ('left', 'center'): (pofst, 0),
                    ('left', 'bottom'): (pofst, pofst)
                }.get((newhalign, newvalign), (0, pofst))
                newofst = Point(newofst)

        else:
            # Align based on position relative to the element
            # Below alignment dictionary works for label on top.
            # Rotate angle for other sides.
            th = theta - label.rotate
            th = {'left': th+90,
                  'bottom': th+180,
                  'right': th+270}.get(label.loc, th)  # type: ignore
            th = (th+360) % 360  # Normalize angle so it's positive, clockwise

            # Alignment for label in different positions
            rotalign: list[tuple[Halign, Valign]] = [
                ('center', 'bottom'),  # label on top
                ('right', 'bottom'),
                ('right', 'center'),   # label on left
                ('right', 'top'),
                ('center', 'top'),     # label on bottom
                ('left', 'top'),
                ('left', 'center'),    # label on right
                ('left', 'bottom')]
            newhalign, newvalign = rotalign[int(round((th/360)*8) % 8)]

            # Ensure label.ofst is a Point (x,y) pair
            if isinstance(label.ofst, (float, int)):
                if label.loc == 'bottom':
                    newofst = (0, -label.ofst)
                elif label.loc == 'left':
                    newofst = (-label.ofst, 0)
                elif label.loc == 'right':
                    newofst = (label.ofst, 0)
                else:
                    newofst = (0, label.ofst)
                newofst = Point(newofst)
        return newhalign, newvalign, newofst

    def _place_label(self, label: Label, theta: float = 0) -> None:
        ''' Adds the label SegmentText to the element, after element placement

            Args:
                label: The Label to add
                theta: Current drawing angle
        '''
        # Make a copy of the label to modify with auto-placement values
        label = Label(label.label, label.loc, label.ofst, label.halign,
                      label.valign, label.rotate, label.fontsize,
                      label.font, label.mathfont, label.color)

        if label.halign is None:
            label.halign = self.params.get('lblalign', (None, None))[0]
        if label.valign is None:
            label.valign = self.params.get('lblalign', (None, None))[1]

        if label.ofst is None:
            label.ofst = self.params.get('lblofst', .1)
        
        label = self._position_label(label, theta)
        halign, valign, ofst = self._align_label(label, theta)
        label.halign = label.halign if label.halign is not None else halign
        label.valign = label.valign if label.valign is not None else valign
        label.ofst = ofst

        # Parameters to send to SegmentText
        segment_params = {
            'color': label.color if label.color else self.params.get('color'),
            'font': label.font if label.font else self.params.get('font'),
            'mathfont': label.mathfont if label.mathfont else self.params.get('mathfont'),
            'fontsize': label.fontsize if label.fontsize else self.params.get('fontsize', 14),
            'align': (label.halign, label.valign),
            'rotation': label.rotate}

        xmax = self.bbox.xmax
        xmin = self.bbox.xmin
        ymax = self.bbox.ymax
        ymin = self.bbox.ymin
        if not math.isfinite(xmax+xmin+ymax+ymin):
            xmax = xmin = ymax = ymin = .1

        if isinstance(label.label, (list, tuple)):
            if label.loc == 'top':
                xdiv = (xmax-xmin)/(len(label.label)+1)
                for i, lbltxt in enumerate(label.label):
                    xy = Point((xmin+xdiv*(i+1), ymax))
                    self.segments.append(SegmentText(xy+label.ofst, lbltxt, **segment_params))
            elif label.loc == 'bottom':
                xdiv = (xmax-xmin)/(len(label.label)+1)
                for i, lbltxt in enumerate(label.label):
                    xy = Point((xmin+xdiv*(i+1), ymin))
                    self.segments.append(SegmentText(xy+label.ofst, lbltxt, **segment_params))
            elif label.loc == 'left':
                ydiv = (ymax-ymin)/(len(label.label)+1)
                for i, lbltxt in enumerate(label.label):
                    xy = Point((xmin, ymin+ydiv*(i+1)))
                    self.segments.append(SegmentText(xy+label.ofst, lbltxt, **segment_params))
            elif label.loc == 'right':
                ydiv = (ymax-ymin)/(len(label.label)+1)
                for i, lbltxt in enumerate(label.label):
                    xy = Point((xmax, ymin+ydiv*(i+1)))
                    self.segments.append(SegmentText(xy+label.ofst, lbltxt, **segment_params))
            elif label.loc == 'center':
                xdiv = (xmax-xmin)/(len(label.label)+1)
                for i, lbltxt in enumerate(label.label):
                    xy = Point((xmin+xdiv*(i+1), 0))
                    self.segments.append(SegmentText(xy+label.ofst, lbltxt, **segment_params))

        elif isinstance(label.label, str):  # keep the elif instead of else for type hinting
            if label.loc and label.loc in self.anchors:
                xy = Point(self.anchors[label.loc])
            elif label.loc == 'top':
                xy = Point(((xmax+xmin)/2, ymax))
            elif label.loc == 'bottom':
                xy = Point(((xmax+xmin)/2, ymin))
            elif label.loc == 'left':
                xy = Point((xmin, (ymax+ymin)/2))
            elif label.loc == 'right':
                xy = Point((xmax, (ymax+ymin)/2))
            elif label.loc == 'center':
                xy = Point(((xmax+xmin)/2, (ymax+ymin)/2))
            else:
                raise ValueError(f'Undefined location {label.loc}')
            self.segments.append(SegmentText(xy+label.ofst, label.label, **segment_params))

    def _draw_on_figure(self):
        ''' Draw the element on a new figure. Useful for _repr_ functions. '''
        if default_canvas.default_canvas == 'matplotlib':
            fig = mplFigure()
        else:
            fig = svgFigure(bbox=self.get_bbox(transform=True))
        if not self._positioned:
            self._place((0, 0), 0)
        fig.set_bbox(self.get_bbox(transform=True))
        self._draw(fig)
        return fig

    def _repr_svg_(self):
        ''' SVG representation for Jupyter '''
        fig = self._draw_on_figure()
        return fig.getimage(ext='svg').decode()

    def _repr_png_(self):
        ''' PNG representation for Jupyter '''
        if default_canvas.default_canvas == 'svg':
            return None
        fig = self._draw_on_figure()
        return fig.getimage(ext='png')

    def _draw(self, fig) -> None:
        ''' Draw the element on a Figure '''
        if len(self.segments) == 0:
            self._place((0, 0), 0)
        for segment in self.segments:
            segment.draw(fig, self.transform, **self.params)

        if self.params.get('elmbbox', False):
            # Draw element bounding box
            bbox = self.get_bbox()
            Segment(((bbox.xmin, bbox.ymin), (bbox.xmax, bbox.ymin),
                     (bbox.xmax, bbox.ymax), (bbox.xmin, bbox.ymax), (bbox.xmin, bbox.ymin)),
                     color='blue', lw=.5).draw(fig, self.transform)


class ElementDrawing(Element):
    ''' Create an element from a Drawing

        Args:
            drawing: The Drawing instance to convert to an element
    '''
    def __init__(self, drawing, **kwargs):
        super().__init__(**kwargs)
        self.drawing = drawing
        self.segments = self.drawing.get_segments()
        self.anchors = self.drawing.anchors
        self.elmparams['drop'] = self.drawing._here
        self.elmparams['d'] = 'right'  # Reset drawing direction
        

class Element2Term(Element):
    ''' Two terminal element. The element leads can be automatically
        extended to the start and ending positions.

        Anchors:
            * start
            * center
            * end
    '''
    _element_defaults = {
        'leadcolor': None,  # Inherit
        'dotradius': 0.075
    }
    def to(self, xy: XY, dx: float = 0, dy: float = 0) -> 'Element2Term':
        ''' Sets ending position of element

            Args:
                xy: Ending position of element
                dx: X-offset from xy position
                dy: Y-offset from xy position
        '''
        xy = Point(xy)
        self._userparams['to'] = Point((xy.x + dx, xy.y + dy))
        return self

    def tox(self, x: float | XY | Element) -> 'Element2Term':
        ''' Sets ending x-position of element (for horizontal elements) '''
        self._userparams['tox'] = x
        return self

    def toy(self, y: float | XY | Element) -> 'Element2Term':
        ''' Sets ending y-position of element (for vertical elements) '''
        self._userparams['toy'] = y
        return self

    def up(self, length: Optional[float] = None) -> 'Element':
        ''' Set the direction to up '''
        if 'd' in self._userparams:
            warnings.warn(f"Duplicated direction parameter in element. `{self._userparams['d']}` changed to `up`.")
        self._userparams['d'] = 'up'
        if length:
            self._userparams['l'] = length
        return self

    def down(self, length: Optional[float] = None) -> 'Element':
        ''' Set the direction to down '''
        if 'd' in self._userparams:
            warnings.warn(f"Duplicated direction parameter in element. `{self._userparams['d']}` changed to `down`.")
        self._userparams['d'] = 'down'
        if length:
            self._userparams['l'] = length
        return self

    def left(self, length: Optional[float] = None) -> 'Element':
        ''' Set the direction to left '''
        if 'd' in self._userparams:
            warnings.warn(f"Duplicated direction parameter in element. `{self._userparams['d']}` changed to `left`.")
        self._userparams['d'] = 'left'
        if length:
            self._userparams['l'] = length
        return self

    def right(self, length: Optional[float] = None) -> 'Element':
        ''' Set the direction to right '''
        if 'd' in self._userparams:
            warnings.warn(f"Duplicated direction parameter in element. `{self._userparams['d']}` changed to `right`.")
        self._userparams['d'] = 'right'
        if length:
            self._userparams['l'] = length
        return self

    def length(self, length: float) -> 'Element2Term':
        ''' Sets total length of element '''
        self._userparams['l'] = length
        return self

    def endpoints(self, start: XY, end: XY) -> 'Element2Term':
        ''' Sets absolute endpoints of element '''
        self._userparams['endpts'] = (start, end)
        return self

    def dot(self, open: bool = False) -> 'Element2Term':
        ''' Add a dot to the end of the element '''
        self._userparams['dot'] = True if not open else 'open'
        return self

    def idot(self, open: bool = False) -> 'Element2Term':
        ''' Add a dot to the input/start of the element '''
        self._userparams['idot'] = True if not open else 'open'
        return self

    def _place(self, dwgxy: XY, dwgtheta: float, **dwgparams) -> tuple[Point, float]:
        ''' Calculate element placement, adding lead extensions '''
        self._dwgparams.clear()
        self._dwgparams.update(dwgparams)
        if not self._positioned:
            self._position()

        totlen = self.params.get('l', self.params.get('unit', 3))
        endpts = self.params.get('endpts', None)
        to = self.params.get('to', None)
        tox = self.params.get('tox', None)
        toy = self.params.get('toy', None)
        anchor = self.params.get('anchor', None)
        zoom = self.params.get('zoom', self.params.get('scale', 1))
        xy = Point(self.params.get('at', dwgxy))

        # set up transformation
        theta = self.params.get('theta', dwgtheta)
        if endpts is not None:
            theta = util.angle(endpts[0], endpts[1])
        elif to is not None:
            theta = util.angle(xy, to)
        elif self.params.get('d') is not None:
            d = self.params.get('d')
            if str(d).lstrip('-').isnumeric():
                theta = d
            else:
                theta = {'u': 90, 'r': 0, 'l': 180, 'd': 270}[d[0].lower()]  # type: ignore

        # Get offset to element position within drawing (global shift)
        if endpts is not None:
            xy = Point(endpts[0])

        if endpts is not None:
            totlen = util.dist(endpts[0], endpts[1])
        elif to is not None:
            # Move until X or Y position is 'end'. Depends on direction
            totlen = util.dist(xy, to)
        elif tox is not None:
            # Allow either full coordinate (only keeping x), or just an x value
            if isinstance(tox, (int, float)):
                x = float(tox)
            else:
                x = tox[0]
            endpt = Point((x, xy[1]))
            totlen = util.dist(xy, endpt)
            theta = 180 if xy.x > x else 0
            self.elmparams['theta'] = theta
        elif toy is not None:
            # Allow either full coordinate (only keeping y), or just a y value
            if isinstance(toy, (int, float)):
                y = toy
            else:
                y = toy[1]
            endpt = Point((xy[0], y))
            totlen = util.dist(xy, endpt)
            theta = -90 if xy.y > y else 90
            self.elmparams['theta'] = theta

        self.anchors['istart'] = self.segments[0].path[0]  # type: ignore
        self.anchors['iend'] = self.segments[0].path[-1]  # type: ignore
        if self.params.get('extend', True):
            assert isinstance(self.segments[0], Segment)
            in_path = self.segments[0].path
            dz = util.delta(in_path[-1], in_path[0])   # Defined delta of path
            in_len = math.hypot(*dz)    # Defined length of path
            try:
                # Zoom is a Point
                lead_len = (totlen - in_len*zoom[0])/2 / zoom[0]
            except TypeError:
                lead_len = (totlen - in_len*zoom)/2 / zoom

            if lead_len > 0:  # Don't make element shorter
                start = Point(in_path[0]) - Point((lead_len, 0))
                end = Point(in_path[-1]) + Point((lead_len, 0))
                self._localshift = -start
                self.segments[0].path = [start] + self.segments[0].path + [end]  # type: ignore
                if in_len > 0 and (leadcolor := self.params.get('leadcolor')):
                    self.segments[0].color = leadcolor

            else:
                start = Point(in_path[0])
                end = Point(in_path[-1])
                self._localshift = Point((0, 0))

            # Adjust position of endpoints (arrowheads, dots, etc.)
            for i, segment in enumerate(self.segments):
                if getattr(segment, 'endref', None) == 'end':
                    xform = Transform(0, end)
                    self.segments[i] = segment.xform(xform)
                elif getattr(segment, 'endref', None) == 'start':
                    xform = Transform(0, start)
                    self.segments[i] = segment.xform(xform)
        else:
            start = Point(self.segments[0].path[0])  # type: ignore
            end = Point(self.segments[0].path[-1])  # type: ignore

        if self.params.get('dot', False):
            fill: Union[bool, str] = 'bg' if self.params['dot'] == 'open' else True
            radius = self.params.get('dotradius', 0.075)
            self.segments.append(SegmentCircle(end, radius=radius, fill=fill, zorder=3))
        if self.params.get('idot', False):
            fill = 'bg' if self.params['idot'] == 'open' else True
            radius = self.params.get('dotradius', 0.075)
            self.segments.append(SegmentCircle(start, radius=radius, fill=fill, zorder=3))

        self._place_anchors(start, end)
        if anchor is not None:
            self._localshift = self._localshift-Point(self.anchors[anchor])
        transform = Transform(theta, xy, self._localshift, zoom=zoom)

        self.absanchors = {}
        if len(self.segments) == 0:
            self.absanchors['start'] = transform.transform((0, 0))
            self.absanchors['end'] = transform.transform((0, 0))
            self.absanchors['center'] = transform.transform((0, 0))
        else:
            self.absanchors['start'] = transform.transform(start)
            self.absanchors['end'] = transform.transform(end)
            self.absanchors['center'] = transform.transform((start+end)/2)

        self.params['drop'] = end
        return super()._place(xy, theta, **dwgparams)

    def _place_anchors(self, start, end):
        ''' Allow positioning anchors before the element is placed '''
        self.anchors['start'] = Point(start)
        self.anchors['end'] = Point(end)
        self.anchors['center'] = (start+end)/2



''' Compound elements made from groups of other elements '''

from __future__ import annotations
from typing import Optional, Sequence, Union

from ..import elements as elm
from ..types import Point
from .. import drawing_stack


class ElementCompound(elm.Element):
    ''' Element onto which other elements can be added like a drawing '''
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.dwgparams = {'unit': kwargs.get('unit', 3),
                          'font': kwargs.get('font', None),
                          'fontsize': kwargs.get('fontsize', 14),
                          'lblofst': kwargs.get('lblofst', .1),
                          'color': kwargs.get('color', 'black'),
                          'lw': kwargs.get('lw', 2),
                          'ls': kwargs.get('ls', '-'),
                          'fill': kwargs.get('fill', 'none')}
        self._here: Point = Point((0, 0))
        self._theta: float = 0
        self.elements: list[elm.Element] = []

        pause_state = drawing_stack.pause
        drawing_stack.pause = True
        self.setup()
        drawing_stack.pause = pause_state

    def __contains__(self, element):
        return element in self.elements
        
    def move_from(self, xy: Point, dx: float = 0, dy: float = 0, theta: Optional[float] = None) -> None:
        ''' Move relative to xy position '''
        xy = Point(xy)
        self._here = Point((xy.x + dx, xy.y + dy))
        if theta is not None:
            self._theta = theta

    def move(self, dx: float = 0, dy: float = 0) -> None:
        ''' Move relative to current position '''
        self._here = Point((self._here.x + dx, self._here.y + dy))

    def add(self, element: elm.Element) -> elm.Element:
        ''' Add an element to the segments list '''
        self.elements.append(element)
        self._here, self._theta = element._place(self._here, self._theta, **self.dwgparams)
        self.segments.extend([s.xform(element.transform, **element.params)
                              for s in element.segments])
        return element


class Optocoupler(ElementCompound):
    ''' Optocoupler element

        Args:
            box: Draw a box around the optocoupler
            boxfill: Color to fill the box
            boxpad: Padding between phototransistor and box
            base: Add a base contact to the phototransistor

        Anchors:
            * anode
            * cathode
            * emitter
            * collector
            * base (if base==True)
    '''
    def __init__(self, box: bool = True, boxfill: str = 'none',
                 boxpad: float = 0.2, base: bool = False, **kwargs):
        self.unit = 1.5
        self.box = box
        self.boxfill = boxfill
        self.boxpad = boxpad
        self.base = base
        super().__init__(unit=self.unit, **kwargs)
    
    def setup(self):
        D = self.add(elm.Diode(d='d'))
        bjt = elm.BjtNpn(d='r', at=(1, -self.unit/2))
        bjt.segments.pop(0)  # Remove base contact
        B = self.add(bjt)
        self.add(elm.Arrow(d='r', at=(.5, -self.unit/2 + .2), l=.6,
                           headwidth=.15, headlength=.2))
        self.add(elm.Arrow(d='r', at=(.5, -self.unit/2 - .2), l=.6,
                           headwidth=.15, headlength=.2))

        bbox = self.get_bbox()
        if self.box:
            self.add(elm.Rect(
                d='r', at=(0, 0),
                corner1=(bbox.xmin-self.boxpad, bbox.ymin-self.boxpad),
                corner2=(bbox.xmax+self.boxpad, bbox.ymax+self.boxpad),
                fill=self.boxfill, zorder=0))

        if self.base:
            self.add(elm.Line(d='l', at=Point((B.get_bbox(transform=True).xmin, D.center[1])), l=.15))  # type: ignore
            E = self.add(elm.Line(d='d', toy=bbox.ymax+self.boxpad))
            self.anchors['base'] = E.end

        A = self.add(elm.Line(d='r', at=bjt.collector, l=self.boxpad))
        B = self.add(elm.Line(d='r', at=bjt.emitter, l=self.boxpad))
        C = self.add(elm.Line(d='l', at=D.start, tox=bbox.xmin-self.boxpad))
        D = self.add(elm.Line(d='l', at=D.end, tox=bbox.xmin-self.boxpad))
        self.anchors['anode'] = C.end
        self.anchors['cathode'] = D.end
        self.anchors['emitter'] = B.end
        self.anchors['collector'] = A.end


class Relay(ElementCompound):
    ''' Relay element with an inductor and switch

        Args:
            unit: Unit length of the inductor
            cycloid: Use cycloid style inductor
            switch: Switch style 'spst', 'spdt', 'dpst', 'dpdt'
            swreverse: Reverse the switch
            swflip: Flip the switch up/down
            core: Show inductor core bar
            link: Show dotted line linking inductor and switch
            box: Draw a box around the relay
            boxfill: Color to fill the box
            boxpad: Spacing between components and box
    '''
    def __init__(self, unit: float = 2, cycl: bool = False, switch: str = 'spst',
                 core: bool = True, box: bool = True, boxfill: str = 'none',
                 boxpad: float = .25, swreverse: bool = False,
                 swflip: bool = False, link: bool = True, **kwargs):
        self.unit = unit
        self.cycl = cycl
        self.switch = switch
        self.core = core
        self.box = box
        self.boxfill = boxfill
        self.boxpad = boxpad
        self.swreverse = swreverse
        self.swflip = swflip
        self.link = link
        super().__init__(unit=unit, **kwargs)

    def setup(self) -> None:
        if self.cycl:
            L = self.add(elm.Inductor2(d='d'))
        else:
            L = self.add(elm.Inductor(d='d'))

        self.anchors['in1'] = L.start
        self.anchors['in2'] = L.end

        Lleft = L.get_bbox(transform=True).xmax
        swleft = Lleft + .6
        if self.core:
            self.add(elm.Line(d='d', at=(Lleft+.15, -self.unit/2),
                              anchor='center', l=1))
            swleft += .1

        SW: Union[elm.Switch, elm.SwitchSpdt2, elm.SwitchDpst, elm.SwitchDpdt]
        if self.switch == 'spst':
            SW = elm.Switch(d='d', reverse=self.swreverse, flip=self.swflip)
            bbox = SW.get_bbox()
            SW._userparams['at'] = (swleft-(-bbox.ymax if self.swflip else bbox.ymin), 0)
            self.add(SW)
            self.anchors['a'] = SW.start
            self.anchors['b'] = SW.end

        elif self.switch == 'spdt':
            SW = elm.SwitchSpdt2(d='d', reverse=self.swreverse, flip=self.swflip)
            bbox = SW.get_bbox()
            SW._userparams['at'] = (swleft-(-bbox.ymax if self.swflip else bbox.ymin), -.5)
            self.add(SW)
            toy = -self.unit if self.swreverse else 0
            a = self.add(elm.Line(d='d' if self.swreverse else 'u', at=SW.a, toy=toy))
            toy = 0 if self.swreverse else -self.unit
            b = self.add(elm.Line(d='u' if self.swreverse else 'd', at=SW.b, toy=toy))
            c = self.add(elm.Line(d='u' if self.swreverse else 'd', at=SW.c, toy=toy))
            self.anchors['a'] = a.end
            self.anchors['b'] = b.end
            self.anchors['c'] = c.end

        elif self.switch == 'dpst':
            SW = elm.SwitchDpst(d='d', link=False, reverse=self.swreverse, flip=self.swflip)
            bbox = SW.get_bbox()
            SW._userparams['at'] = (swleft-(-bbox.ymax if self.swflip else bbox.ymin), -.5)
            self.add(SW)
            toy = 0 if self.swreverse else -self.unit
            t1 = self.add(elm.Line(d='u' if self.swreverse else 'd', at=SW.t1, toy=toy))
            t2 = self.add(elm.Line(d='u' if self.swreverse else 'd', at=SW.t2, toy=toy))
            toy = -self.unit if self.swreverse else 0
            p1 = self.add(elm.Line(d='d' if self.swreverse else 'u', at=SW.p1, toy=toy))
            p2 = self.add(elm.Line(d='d' if self.swreverse else 'u', at=SW.p2, toy=toy))
            self.anchors['t1'] = t1.end
            self.anchors['t2'] = t2.end
            self.anchors['p1'] = p1.end
            self.anchors['p2'] = p2.end

        elif self.switch == 'dpdt':
            SW = elm.SwitchDpdt(d='d', link=False, reverse=self.swreverse, flip=self.swflip)
            bbox = SW.get_bbox()
            SW._userparams['at'] = (swleft-(-bbox.ymax if self.swflip else bbox.ymin), -.5)
            self.add(SW)
            toy = 0 if self.swreverse else -self.unit
            t1 = self.add(elm.Line(d='u' if self.swreverse else 'd', at=SW.t1, toy=toy))
            t2 = self.add(elm.Line(d='u' if self.swreverse else 'd', at=SW.t2, toy=toy))
            t3 = self.add(elm.Line(d='u' if self.swreverse else 'd', at=SW.t3, toy=toy))
            t4 = self.add(elm.Line(d='u' if self.swreverse else 'd', at=SW.t4, toy=toy))
            toy = -self.unit if self.swreverse else 0
            p1 = self.add(elm.Line(d='d' if self.swreverse else 'u', at=SW.p1, toy=toy))
            p2 = self.add(elm.Line(d='d' if self.swreverse else 'u', at=SW.p2, toy=toy))
            self.anchors['t1'] = t1.end
            self.anchors['t2'] = t2.end
            self.anchors['t3'] = t3.end
            self.anchors['t4'] = t4.end
            self.anchors['p1'] = p1.end
            self.anchors['p2'] = p2.end

        if self.link:
            tox = SW.get_bbox().ymax if self.swflip else SW.get_bbox().ymin
            self.add(elm.Line(d='r', at=(Lleft+.2, -self.unit/2), ls=':', tox=tox))

        if self.box:
            bbox = self.get_bbox()
            self.add(elm.Rect(d='r', at=(0, 0),
                              corner1=(bbox.xmin-self.boxpad, bbox.ymin+.2),
                              corner2=(bbox.xmax+self.boxpad, bbox.ymax-.2),
                              fill=self.boxfill, zorder=0))


class Wheatstone(ElementCompound):
    ''' Wheatstone Resistor Bridge

        Args:
            vout: draw output terminals inside the bridge
            labels: Labels to draw on each resistor

        Anchors:
            * N
            * S
            * E
            * W
            * vo1 (if vout==True)
            * vo2 (if vout==True)
    '''
    def __init__(self, vout: bool = False, labels: Optional[Sequence[str]] = None, **kwargs):
        self.vout = vout
        self.labels = labels
        super().__init__(**kwargs)

    def setup(self):
        A = elm.Resistor().theta(45)
        B = elm.Resistor().theta(-45)
        C = elm.Resistor().theta(-135)
        D = elm.Resistor().theta(135)

        locs = ['top', 'top', 'bottom', 'bottom']
        if self.labels:
            for label, r, loc in zip(self.labels, [A, B, C, D], locs):
                r.label(label, loc)

        self.add(A)
        self.add(B)
        self.add(C)
        self.add(D)
        if self.vout:
            self.add(elm.Line().right().at(A.start).length(1))
            vo1 = self.add(elm.Dot(open=True))
            self.add(elm.Line().left().at(C.start).length(1))
            vo2 = self.add(elm.Dot(open=True))
            self.anchors['vo1'] = vo1.center
            self.anchors['vo2'] = vo2.center

        self.anchors['W'] = A.start
        self.anchors['N'] = B.start
        self.anchors['E'] = C.start
        self.anchors['S'] = D.start
        self.params['theta'] = 0


class Rectifier(ElementCompound):
    ''' Diode Rectifier Bridge

        Args:
            fill: Fill the didoes
            labels: Labels to draw on each resistor

        Anchors:
            * N
            * S
            * E
            * W
    '''
    def __init__(self, fill=False, labels=None, **kwargs):
        self.fill = fill
        self.labels = labels
        super().__init__(**kwargs)

    def setup(self):
        A = elm.Diode(fill=self.fill).theta(45)
        B = elm.Diode(fill=self.fill).theta(-45)
        C = elm.Diode(fill=self.fill).theta(-135).reverse()
        D = elm.Diode(fill=self.fill).theta(135).reverse()

        locs = ['top', 'top', 'bottom', 'bottom']
        if self.labels:
            for label, d, loc in zip(self.labels, [A, B, C, D], locs):
                d.label(label, loc)

        self.add(A)
        self.add(B)
        self.add(C)
        self.add(D)
        self.anchors['W'] = A.start
        self.anchors['N'] = B.start
        self.anchors['E'] = C.start
        self.anchors['S'] = D.start
        self.params['theta'] = 0



''' Affine transformation matrix operations for SVG `transform` attributes '''
from __future__ import annotations
from typing import Sequence, Optional, Tuple
import math

from ..util import Point


Matrix3x3 = Tuple[Tuple[float, float, float], Tuple[float, float, float], Tuple[float, float, float]]


def matrix_translate(dx: float = 0, dy: float = 0) -> Matrix3x3:
    ''' Translation matrix '''
    return ((1., 0., dx),
            (0., 1., dy),
            (0., 0., 1.))


def matrix_rotate(theta: float = 0, cx: float = 0, cy: float = 0) -> Matrix3x3:
    ''' Rotation matrix, theta in degrees, about center (cx, cy) '''
    costh = math.cos(math.radians(theta))
    sinth = math.sin(math.radians(theta))
    return ((costh, -sinth, -cx*costh + cy*sinth + cx),
            (sinth, costh, -cx*sinth - cy*costh + cy),
            (0., 0., 1.))


def matrix_scale(kx: float = 1., ky: Optional[float] = None) -> Matrix3x3:
    ''' Scale matrix '''
    if ky is None:
        ky = kx
    return ((kx, 0., 0.),
            (0., ky, 0.),
            (0., 0., 1.))


def matrix_skewx(k: float = 1.) -> Matrix3x3:
    ''' X-Skew matrix '''
    return ((1., k,  0.),
            (0., 1., 0.),
            (0., 0., 1.))


def matrix_skewy(k: float = 1.) -> Matrix3x3:
    ''' Y-Skew matrix '''
    return ((1., 0., 0.),
            (k,  1., 0.),
            (0., 0., 1.))


def matrix(a: float = 1., b: float = 0., c: float = 0., d: float = 1., e: float = 0, f: float = 0) -> Matrix3x3:
    ''' Full matrix as defined by SVG transform '''
    return ((a, c, e),
            (b,  d, f),
            (0., 0., 1.))


def transform(point: Point, xform: Matrix3x3) -> Point:
    ''' Apply transform to point (matrix multiply) '''
    pt = [point[0], point[1], 1.]
    x = pt[0] * xform[0][0] + pt[1] * xform[0][1] + pt[2]*xform[0][2]
    y = pt[0] * xform[1][0] + pt[1] * xform[1][1] + pt[2]*xform[1][2]
    # z = pt[0] * xform[2][0] + pt[1] * xform[2][1] + pt[2]*xform[2][2]
    # assert math.isclose(z, 1)
    return Point((x, y))


def transform_all(pt, xform: Sequence[Matrix3x3]) -> Point:
    ''' Apply series of transforms to the point '''
    for xf in xform:
        pt = transform(pt, xf)
    return pt



''' Signal processing elements '''

from __future__ import annotations
import math
from typing import Optional, Sequence, Literal

from ..util import linspace, Point
from ..segments import Segment, SegmentCircle, SegmentText, SegmentArc
from ..elements import Element
from ..types import XY


FilterType = Literal['lp', 'bp', 'hp', 'notch']


class Square(Element):
    ''' Empty square element

        Anchors:
            * N
            * S
            * E
            * W
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), (0, .5), (1, .5),
                                      (1, -.5), (0, -0.5), (0, 0)]))
        self.elmparams['lblloc'] = 'center'
        self.elmparams['lblofst'] = 0
        self.anchors['N'] = (0.5, 0.5)
        self.anchors['S'] = (0.5, -0.5)
        self.anchors['E'] = (1, 0)
        self.anchors['W'] = (0, 0)
        self.elmparams['drop'] = (1, 0)


class Circle(Element):
    ''' Empty circle element

        Anchors:
            * N
            * S
            * E
            * W
            * NW
            * NE
            * SW
            * SE
    '''
    _element_defaults = {
        'radius': 0.5
    }
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        rad = self.params['radius']
        k = rad*math.sqrt(2)/2  # Diagonal distance
        self.segments.append(SegmentCircle((rad, 0), rad))
        self.elmparams['lblloc'] = 'center'
        self.elmparams['lblofst'] = 0
        self.elmparams['drop'] = (2*rad, 0)
        self.anchors['N'] = (rad, rad)
        self.anchors['S'] = (rad, -rad)
        self.anchors['E'] = (2*rad, 0)
        self.anchors['W'] = (0, 0)
        self.anchors['NW'] = (rad-k, k)
        self.anchors['NE'] = (rad+k, k)
        self.anchors['SW'] = (rad-k, -k)
        self.anchors['SE'] = (rad+k, -k)
        self.anchors['center'] = (rad, 0)


class Sum(Circle):
    ''' Summation element (+ symbol)

        Anchors:
            * N
            * S
            * E
            * W
            * NW
            * NE
            * SW
            * SE
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(.5, .2), (.5, -.2)]))
        self.segments.append(Segment([(.3, 0), (.7, 0)]))


class SumSigma(Circle):
    ''' Summation element (Greek Sigma symbol)

        Anchors:
            * N
            * S
            * E
            * W
            * NW
            * NE
            * SW
            * SE
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(SegmentText(label=r'$\Sigma$',
                                         pos=(0.45, 0),
                                         align=('center', 'center')))


class Mixer(Circle):
    ''' Mixer

        Args:
            N: text in North sector
            S: text in South sector
            E: text in East sector
            W: text in West sector
            font: Font family/name
            fontsize: Point size of label font

        Anchors:
            * N
            * S
            * E
            * W
            * NW
            * NE
            * SW
            * SE
    '''
    def __init__(self, N: Optional[str] = None, E: Optional[str] = None, S: Optional[str] = None, W: Optional[str] = None,
                 font: Optional[str] = None, fontsize: float = 10, **kwargs):
        super().__init__(**kwargs)
        rad = .5
        k = rad*math.sqrt(2)/2  # Diagonal distance
        self.segments.append(Segment([(rad+k, k), (rad-k, -k)], lw=1))
        self.segments.append(Segment([(rad+k, -k), (rad-k, k)], lw=1))
        self.elmparams['lblloc'] = 'top'
        self.elmparams['lblofst'] = 0.2
        if N:
            self.segments.append(SegmentText(
                (rad, rad/2+0.01), N,
                font=font, fontsize=fontsize))
        if S:
            self.segments.append(SegmentText(
                (rad, -rad/2-0.065), S,
                font=font, fontsize=fontsize))
        if E:
            self.segments.append(SegmentText(
                (rad*1.5+0.05, 0), E,
                font=font, fontsize=fontsize))
        if W:
            self.segments.append(SegmentText(
                (rad/2-0.05, 0), W,
                font=font, fontsize=fontsize))


class Speaker(Element):
    ''' Speaker with only one terminal '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), (0, 0.25), (0.25, 0.25),
                                      (0.25, -.25), (0, -.25), (0, 0)]))
        self.segments.append(Segment([(0.25, 0.25), (0.5, 0.5), (0.5, -0.5),
                                      (0.25, -0.25), (.25, .25)]))


class Amp(Element):
    ''' Amplifier

        Anchors:
            * in
            * out
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        amph = 1.
        ampl = .75
        self.segments.append(Segment([(0, 0), (0, -amph/2), (ampl, 0),
                                      (0, amph/2), (0, 0)]))
        self.elmparams['drop'] = (ampl, 0)
        self.anchors['input'] = (0, 0)
        self.anchors['out'] = (ampl, 0)


def _makesine() -> Sequence[XY]:
    sinx = linspace(-math.pi, math.pi, num=20)
    siny = [-math.sin(x)/10 for x in sinx]
    sinx = [x / math.pi * .3 + .5 for x in sinx]
    path = list(zip(sinx, siny))
    return path


class OscillatorBox(Square):
    ''' Oscillator in a square

        Anchors:
            * N
            * S
            * E
            * W
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        path = _makesine()
        self.segments.append(Segment(path))


class Oscillator(Circle):
    ''' Oscillator in a circle

        Anchors:
            * N
            * S
            * E
            * W
            * NW
            * NE
            * SW
            * SE
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        path = _makesine()
        self.segments.append(Segment(path))


class Filter(Square):
    ''' Filter

        Args:
            response: Filter response ('lp', 'bp', 'hp', or 'notch') for
                low-pass, band-pass, high-pass, and notch/band-stop filters

        Anchors:
            * N
            * S
            * E
            * W
    '''
    def __init__(self, response: Optional[FilterType] = None, **kwargs):
        super().__init__(**kwargs)
        path = _makesine()
        path1 = [Point((p[0], p[1]+.25)) for p in path]
        path2 = [Point((p[0], p[1]-.25)) for p in path]
        self.segments.append(Segment(path))
        self.segments.append(Segment(path1))
        self.segments.append(Segment(path2))

        if response:
            if response.lower() in ['bp', 'lp']:
                # Slash through high f
                self.segments.append(Segment([(.45, .17), (.55, .33)]))
            if response.lower() in ['bp', 'hp']:
                # Slash through low f
                self.segments.append(Segment([(.45, -.33), (.55, -.17)]))
            if response.lower() in ['lp', 'hp', 'notch']:
                # Slash through mid f
                self.segments.append(Segment([(.45, -.08), (.55, .08)]))


class Adc(Element):
    ''' Analog to digital converter

        Anchors:
            * in
            * out
            * E (same as in)
            * W (same as out)
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), (.22, .5), (1.4, .5), (1.4, -.5),
                                      (.22, -.5), (0, 0)]))
        self.elmparams['lblloc'] = 'center'
        self.elmparams['lblofst'] = 0
        self.elmparams['drop'] = (1.4, 0)
        self.anchors['input'] = (0, 0)
        self.anchors['out'] = (1.4, 0)
        self.anchors['E'] = (1.4, 0)
        self.anchors['W'] = (0, 0)


class Dac(Element):
    ''' Digital to analog converter

        Anchors:
            * in
            * out
            * E (same as in)
            * W (same as out)
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), (0, .5), (1.18, .5), (1.4, 0),
                                      (1.18, -.5), (0, -.5), (0, 0)]))
        self.elmparams['lblloc'] = 'center'
        self.elmparams['lblofst'] = 0
        self.elmparams['drop'] = (1.4, 0)
        self.anchors['input'] = (0, 0)
        self.anchors['out'] = (1.4, 0)
        self.anchors['E'] = (1.4, 0)
        self.anchors['W'] = (0, 0)


class Demod(Square):
    ''' Demodulator (box with a diode in it)

        Anchors:
            * N
            * S
            * E
            * W
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(.15, 0), (.3, 0)]))
        self.segments.append(Segment([(.3, .25), (.7, 0), (.3, -.25), (.3, .25)]))
        self.segments.append(Segment([(.7, .25), (.7, -.25)]))
        self.segments.append(Segment([(.7, 0), (.85, 0)]))


class Circulator(Circle):
    ''' Circulator (circle with an arrow in it)

        Anchors:
            * N
            * S
            * E
            * W
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        radius = 0.5
        self.segments.append(
            SegmentArc((0.5, 0), 0.9 * radius, 0.9 * radius, 20, 200, "cw"))


class Isolator(Square):
    ''' Isolator (box with an arrow in it)

        Anchors:
            * N
            * S
            * E
            * W
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        margin = 0.25
        path = ((margin, 0), (1 - margin, 0))
        self.segments.append(Segment(path, arrow="->"))


class VGA(Amp):
    ''' Variable Gain Amplifier (amplifier symbol with an arrow over it)

        Args:
            tuneup: Set tune above or below the symbol

        Anchors:
            * input
            * out
            * tune
    '''
    def __init__(self, tuneup: bool = True, **kwargs):
        super().__init__(**kwargs)
        path = ((-0.1, -0.5), (0.75, 0.5))
        self.segments.append(Segment(path, arrow="->"))
        self.anchors['tune'] = (0.325, 0.5 if tuneup else -0.5)



''' Elements based on images '''
from __future__ import annotations
from typing import Optional, BinaryIO

from .elements import Element
from ..segments import SegmentImage
from ..util import Point


class ElementImage(Element):
    ''' Element from an Image file

        Args:
            image: Image filename or open file pointer
            width: Width to draw image in Drawing
            height: Height to draw image in Drawing
            xy: Origin (lower left corner)
    '''
    def __init__(self, image: str | BinaryIO,
                 width: float,
                 height: float,
                 xy: Point = Point((0, 0)),
                 imgfmt: Optional[str] = None,
                 **kwargs):
        super().__init__(**kwargs)
        zorder = kwargs.get('zorder', 1)
        self.segments.append(SegmentImage(image=image, xy=xy, width=width, height=height,
                                          zorder=zorder, imgfmt=imgfmt))
        self.elmparams['theta'] = 0



from .schemdraw import Drawing, use, config, theme, debug
from .segments import Segment, SegmentCircle, SegmentArc, SegmentText, SegmentPoly, SegmentBezier, SegmentPath
from .transform import Transform
from .types import ImageFormat
from .backends.svg import config as svgconfig
from .backends.svg import settextmode

__all__ = [
    "Drawing", "use", "config", "theme", "debug", "Segment", "SegmentCircle", "SegmentArc", "SegmentText",
    "SegmentPath",
    "SegmentPoly", "SegmentBezier", "Transform", "ImageFormat", "settextmode", "svgconfig"
]

__version__ = '0.19'



''' Cable elements, coaxial and triaxial '''
from __future__ import annotations
from typing import Optional
import math
import warnings

from ..segments import Segment, SegmentArc
from .elements import Element2Term, gap


class Coax(Element2Term):
    ''' Coaxial cable element.

        Args:
            length: Total length of the cable, excluding lead extensions. [default: 3]
            radius: Radius of shield [default: 0.3]
            leadlen: Distance (x) from start of center conductor to
                start of shield. [default: 0.6]

        Anchors:
            * shieldstart
            * shieldstart_top
            * shieldend
            * shieldend_top
            * shieldcenter
            * shieldcenter_top
    '''
    _element_defaults = {
        'length': 3,
        'leadlen': 0.6,
        'radius': 0.3,
    }
    def __init__(self, *,
                 length: Optional[float] = None,
                 radius: Optional[float] = None,
                 leadlen: Optional[float] = None,
                 **kwargs):
        super().__init__(**kwargs)
        _leadlen = self.params['leadlen']
        _length = self.params['length']
        _radius = self.params['radius']
        self.segments.append(Segment(    # Center conductor
                [(0, 0), (_leadlen, 0), gap, (_length-_leadlen+_radius/2, 0),
                 (_length, 0)]))
        self.segments.append(Segment(
            [(_leadlen, _radius), (_length-_leadlen, _radius)]))   # Top
        self.segments.append(Segment(
            [(_leadlen, -_radius), (_length-_leadlen, -_radius)]))  # Bottom
        self.segments.append(SegmentArc(
            (_leadlen, 0), width=_radius, height=_radius*2, theta1=0, theta2=360))
        self.segments.append(SegmentArc(
            (_length-_leadlen, 0), width=_radius,
            height=_radius*2, theta1=270, theta2=90))
        self.anchors['shieldstart'] = (_leadlen, -_radius)
        self.anchors['shieldstart_top'] = (_leadlen, _radius)
        self.anchors['shieldend'] = (_length-_leadlen, -_radius)
        self.anchors['shieldend_top'] = (_length-_leadlen, _radius)
        self.anchors['shieldcenter'] = (_length/2, -_radius)
        self.anchors['shieldcenter_top'] = (_length/2, _radius)

        if _radius/2 > _leadlen:
            warnings.warn('Coax leadlen < radius/2. Coax may be malformed.')
        if _leadlen*2 > _length:
            warnings.warn('Coax length < 2*leadlen. Coax may be malformed.')


class Triax(Element2Term):
    ''' Triaxial cable element.

        Args:
            length: Total length of the cable [default: 3]
            radiusinner: Radius of inner guard [default: 0.3]
            radiusouter: Radius of outer shield [default: 0.6]
            leadlen: Distance (x) from start of center conductor to
                start of guard. [default: 0.6]
            shieldofststart: Distance from start of inner guard to start
                of outer shield [default: 0.3]
            shieldofstend: Distance from end of outer shield to end
                of inner guard [default: 0.3]

        Anchors:
            * shieldstart
            * shieldstart_top
            * shieldend
            * shieldend_top
            * shieldcenter
            * shieldcenter_top
            * guardstart
            * guardstart_top
            * guardend
            * guardend_top
    '''
    _element_defaults = {
        'length': 3,
        'leadlen': 0.6,
        'radiusinner': 0.3,
        'radiusouter': 0.6,
        'shieldofststart': 0.3,
        'shieldofstend': 0.3
    }
    def __init__(self, *,
                 length: Optional[float] = None,
                 leadlen: Optional[float] = None,
                 radiusinner: Optional[float] = None,
                 radiusouter: Optional[float] = None,
                 shieldofststart: Optional[float] = None,
                 shieldofstend: Optional[float] = None,
                 **kwargs):
        super().__init__(**kwargs)
        r1 = self.params['radiusinner']
        r2 = self.params['radiusouter']
        _length = self.params['length']
        _leadlen = self.params['leadlen']
        _shieldofstend = self.params['shieldofstend']
        _shieldofststart = self.params['shieldofststart']

        if r2 < r1:
            raise ValueError('Triax inner radius > outer radius')

        xshield = r2/2 * math.sqrt(1 - r1**2/r2**2)
        if _shieldofststart - xshield > -r1/2:
            thetashield = 180 - math.degrees(math.atan2(r1, xshield))
        else:
            thetashield = 180

        if _length - _leadlen - _shieldofstend + r2 < _length:
            # Include the inner guard on output side
            self.segments.append(Segment(  # Center conductor (first)
                [(0, 0), (_leadlen, 0), gap,
                 (_length-_leadlen+r1/2, 0),
                 (_length, 0)]))
            self.segments.append(Segment(
                [(_length-_leadlen-_shieldofstend+xshield, r1),
                 (_length-_leadlen, r1)]))  # guard (inner) top/right
            self.segments.append(Segment(
                [(_length-_leadlen-_shieldofstend+xshield, -r1),
                 (_length-_leadlen, -r1)]))  # guard (inner) bottom/right
            self.segments.append(SegmentArc(
                (_length-_leadlen, 0), width=r1, height=r1*2,
                theta1=270, theta2=90))
        else:
            # Don't include inner guard on output side
            self.segments.append(Segment(    # Center conductor
                [(0, 0), (_leadlen, 0), gap,
                 (_length-_leadlen-_shieldofstend+r2/2, 0)]))

        # Start with shapes that are always shown...
        self.segments.append(Segment(  # guard (inner) top/left
            [(_leadlen, r1), (_leadlen+_shieldofststart+xshield, r1)]))
        self.segments.append(Segment(  # guard (inner) bottom/left
            [(_leadlen, -r1), (_leadlen+_shieldofststart+xshield, -r1)]))
        self.segments.append(Segment(  # shield (outer) top
            [(_leadlen+_shieldofststart, r2),
             (_length-_leadlen-_shieldofstend, r2)]))
        self.segments.append(Segment(  # shield (outer) bottom
            [(_leadlen+_shieldofststart, -r2),
             (_length-_leadlen-_shieldofstend, -r2)]))

        self.segments.append(SegmentArc(
            (_leadlen, 0), width=r1, height=r1*2,
            theta1=0, theta2=360))
        self.segments.append(SegmentArc(
            (_leadlen+_shieldofststart, 0), width=r2,
            height=r2*2, theta1=-thetashield, theta2=thetashield))
        self.segments.append(SegmentArc(
            (_length-_leadlen-_shieldofstend, 0), width=r2,
            height=r2*2, theta1=270, theta2=90))

        self.anchors['guardstart'] = (_leadlen, -r1)
        self.anchors['guardstart_top'] = (_leadlen, r1)
        self.anchors['guardend'] = (_length-_leadlen, -r1)
        self.anchors['guardend_top'] = (_length-_leadlen, r1)
        self.anchors['shieldstart'] = (_leadlen+_shieldofststart, -r2)
        self.anchors['shieldstart_top'] = (_leadlen+_shieldofststart, r2)
        self.anchors['shieldend'] = (_length-_leadlen-_shieldofstend, -r2)
        self.anchors['shieldend_top'] = (_length-_leadlen-_shieldofstend, r2)
        self.anchors['shieldcenter'] = (_length/2, -r2)
        self.anchors['shieldcenter_top'] = (_length/2, r2)

        if r2 <= r1:
            warnings.warn('Triax outer radius < inner radius')

        if _leadlen+_shieldofststart > _length-_leadlen-_shieldofstend:
            warnings.warn('Triax too short for outer radius')
